###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Thin wrapper around Google Sevice API for integraton into StarThinker.

This does not change or augment the standard API calls other than the following:

  * Allows passing of auth parameter to constructor, required for switching.
  * Execute statement is overloaded to include iterator for responses with
  nextPageToken.
  * Retries handle some common errors and have a back off scheme.
  * JSON based configuration allows StarThinker recipe definitions.
  * Pre-defined functions for each API can be added to fix version and uri
  options.

"""

import json
import traceback
import httplib2
from time import sleep
from googleapiclient.errors import HttpError
from googleapiclient.discovery import Resource
from ssl import SSLError

try:
  import httplib
except:
  import http.client as httplib

from starthinker.util.auth import get_service
from starthinker.util.project import project

RETRIABLE_EXCEPTIONS = (httplib2.HttpLib2Error, IOError, httplib.NotConnected,
                        httplib.IncompleteRead, httplib.ImproperConnectionState,
                        httplib.CannotSendRequest, httplib.CannotSendHeader,
                        httplib.ResponseNotReady, httplib.BadStatusLine)

RETRIABLE_STATUS_CODES = [500, 502, 503, 504]


def API_Retry(job, key=None, retries=3, wait=31):
  """ API retry that includes back off and some common error handling.

  CAUTION:  Total timeout cannot exceed 5 minutes or the SSL token expires for
  all future calls.

  For critical but recoverable errors, the back off executes [retry] times.
  Each time the [wait] is doubled.
  By default retries are: 0:31 + 1:02 + 2:04 = 3:37 ( minutes )
  The recommended minimum wait is 60 seconds for most APIs.

  * Errors retried: 429, 500, 503
  * Errors ignored: 409 - already exists ( triggered by create only and also
  returns None )
  * Errors raised: ALL OTHERS

  Args:
    * job: (object) Everything before the execute() statement.
    * key: (string) key of value from data to return.
    * retries: (int) Number of times to try the job.
    * wait: (seconds) Time to wait in seconds between retries.

  Returns:
    * JSON result of job or key value from JSON result if job succeed.
    * None if object already exists.

  Raises:
    * Any exceptions not listed in comments above.

  """

  try:
    # try to run the job and return the response
    data = job.execute()
    return data if not key else data.get(key, [])

  # API errors
  except HttpError as e:
    # errors that can be overcome or re-tried ( 403 is rate limit and others, needs deep dive )
    if e.resp.status in [403, 409, 429, 500, 503]:
      content = json.loads(e.content.decode())
      # already exists ( ignore benign )
      if content['error']['code'] == 409:
        return None
      # permission denied ( won't change on retry so raise )
      elif content.get(
          'error', {}).get('status') == 'PERMISSION_DENIED' or content.get(
              'error', {}).get('errors', [{}])[0].get('reason') == 'forbidden':
        raise
      elif retries > 0:
        if project.verbose:
          print('API ERROR:', str(e))
        if project.verbose:
          print('API RETRY / WAIT:', retries, wait)
        sleep(wait)
        return API_Retry(job, key, retries - 1, wait * 2)
      # if no retries, raise
      else:
        raise
    # raise all other errors that cannot be overcome
    else:
      raise

  # HTTP transport errors
  except RETRIABLE_EXCEPTIONS as e:
    if retries > 0:
      if project.verbose:
        print('HTTP ERROR:', str(e))
      if project.verbose:
        print('HTTP RETRY / WAIT:', retries, wait)
      sleep(wait)
      return API_Retry(job, key, retries - 1, wait * 2)
    else:
      raise

  # SSL timeout errors
  except SSLError as e:
    # most SSLErrors are not retriable, only timeouts, but
    # SSLError has no good error type attribute, so we search the message
    if retries > 0 and 'timed out' in e.message:
      if project.verbose:
        print('SSL ERROR:', str(e))
      if project.verbose:
        print('SSL RETRY / WAIT:', retries, wait)
      sleep(wait)
      return API_Retry(job, key, retries - 1, wait * 2)
    else:
      raise


def API_Iterator(function, kwargs, results=None):
  """ See below API_Iterator_Instance for documentaion, this is just an iter wrapper.

      Returns:
        iter(API_Iterator_Instance(function, kwargs, results))
  """

  class API_Iterator_Instance():
    """A helper class that iterates multiple results, automatically called by execute.

      This is a standard python iterator definition, no need to document
      functions.

      The only job this has is to handle Google API iteration, as such it can be
      called
      on any API call that reurns a 'nextPageToken' in the result.

      For example if calling the DCM list placement API:

        https://developers.google.com/doubleclick-advertisers/v3.3/placements/list

        function = get_service('dfareporting', 'v3.3', 'user').placements().list
        kwargs = { 'profile_id':1234, 'archived':False }
        for placement in API_Iterator(function, kwargs):
          print(placement)

      Can be called independently but automatically built into API...execute()
      so
      use that instead.

      Args:
        function: (function) function of API call to iterate, definiton not
          instance. kwargs" (dict) arguments to be passed to the function on
          each fetch results (json) optional, the first set of results given (
          if already fetched )

    Returns:
      Iterator over JSON objects.
    """

    def __init__(self, function, kwargs, results=None):
      self.function = function
      self.kwargs = kwargs
      self.results = results
      self.position = 0
      self.iterable = None
      self.__find_tag__()

    def __find_tag__(self):
      # find the only list item for a paginated response, JOSN will only have list type, so ok to be specific
      if self.results:  # None and {} both excluded
        for tag in iter(self.results.keys()):
          if isinstance(self.results[tag], list):
            self.iterable = tag
            break

        # this shouldn't happen but some APIs simply omit the key if no results
        if self.iterable is None and project.verbose:
          print('WARNING API RETURNED NO KEYS WITH LISTS:',
                ', '.join(self.results.keys()))

    def __iter__(self):
      return self

    def __next__(self):
      return self.next()

    def next(self):

      # if no initial results, get some, empty results {} different
      if self.results is None:
        self.results = API_Retry(self.function(**self.kwargs))
        self.__find_tag__()

      # if empty results or exhausted page, get next page
      if self.iterable and self.position >= len(self.results[self.iterable]):
        page_token = self.results.get('nextPageToken', None)
        if page_token:

          if 'body' in self.kwargs:
            self.kwargs['body']['pageToken'] = page_token
          else:
            self.kwargs['pageToken'] = page_token

          self.results = API_Retry(self.function(**self.kwargs))
          self.position = 0

        else:
          raise StopIteration

      # if results remain, return them
      if self.iterable and self.position < len(self.results[self.iterable]):
        value = self.results[self.iterable][self.position]
        self.position += 1
        return value

      # if pages and results exhausted, stop
      else:
        raise StopIteration

  return iter(API_Iterator_Instance(function, kwargs, results))


class API():
  """A wrapper around Google API with built in helpers for StarThinker.

    The wrapper mimics function calls, storing the m in a stack, until it
    encounters
    execute().  Then it uses the stored stack and arguments to call the actual
    API.
    This allows handlers on execute such as API_Retry and API_Iterator.

    See module level description for wrapped changes to Google API.  The class
    is
    designed to be a connector to JSON, hence the configuraton is a JSON object.

    configuration = {
      "api":"doubleclickbidmanager",
      "version":"v1",
      "auth":"user",
      "iterate":False
    }
    api = API(configuration).placements().list(profile_id=1234,
    archived=False).execute()

    Args:
      configuration: (json) see example above, configures all API parameters

    Returns:
      If nextpageToken in result or iterate is True: return iterator of API
      response
      Otherwise: returns API response
  """

  def __init__(self, configuration):
    self.api = configuration['api']
    self.version = configuration['version']
    self.auth = configuration['auth']
    self.uri = configuration.get('uri', None)
    self.key = configuration.get('key', None)
    self.function_stack = list(
        filter(None,
               configuration.get('function', '').split('.')))
    self.function_kwargs = configuration.get('kwargs', {})
    self.iterate = configuration.get('iterate', False)
    self.headers = configuration.get('headers', {})

    self.function = None
    self.job = None
    self.response = None

  # for debug purposes
  def __str__(self):
    return '%s.%s.%s' % (self.api, self.version, '.'.join(self.function_stack))

  # builds API function stack
  def __getattr__(self, function_name):
    self.function_stack.append(function_name)

    def function_call(**kwargs):
      self.function_kwargs = kwargs
      return self

    return function_call

  # for calling function via string ( chain using dot notation )
  def call(self, function_chain):
    for function_name in function_chain.split('.'):
      self.function_stack.append(function_name)
    return self

  # matches API execute with built in iteration and retry handlers
  def execute(self, run=True, iterate=True):
    # start building call sequence with service object
    self.function = get_service(
        api=self.api,
        version=self.version,
        auth=self.auth,
        headers=self.headers,
        key=self.key,
        uri_file=self.uri)

    # build calls along stack
    # do not call functions, as the abstract is necessary for iterator page next calls
    for f_n in self.function_stack:
      #print(type(self.function), isinstance(self.function, Resource))
      self.function = getattr(
          self.function
          if isinstance(self.function, Resource) else self.function(), f_n)

    # for cases where job is handled manually, save the job
    self.job = self.function(**self.function_kwargs)

    if run:
      self.response = API_Retry(self.job)

      # if paginated, automatically iterate
      if (iterate and (self.iterate or (isinstance(self.response, dict) and
                                        'nextPageToken' in self.response))):
        return API_Iterator(self.function, self.function_kwargs, self.response)

      # if not paginated, return object as is
      else:
        return self.response

    # if not run, just return job object ( for chunked upload for example )
    else:
      return self.job

  def upload(self, retries=5, wait=61):
    job = self.execute(run=False)
    response = None

    while response is None:
      error = None

      try:
        print('Uploading file...')
        status, response = job.next_chunk()
        if 'id' in response:
          print("Object id '%s' was successfully uploaded." % response['id'])
        else:
          exit('The upload failed with an unexpected response: %s' % response)

      except HttpError as e:
        if retries > 0 and e.resp.status in RETRIABLE_STATUS_CODES:
          error = 'A retriable HTTP error %d occurred:\n%s' % (
              e.resp.status, e.content.decode())
        else:
          raise

      except RETRIABLE_EXCEPTIONS as e:
        if retries > 0:
          error = 'A retriable error occurred: %s' % e
        else:
          raise

      if error is not None:
        print(error)
        retries -= 1
        wait = wait * 2
        print('Sleeping %d seconds and then retrying...' % wait)
        time.sleep(wait)


def API_BigQuery(auth, iterate=False):
  """BigQuery helper configuration for Google API.

  Defines agreed upon version.
  """

  configuration = {
      'api': 'bigquery',
      'version': 'v2',
      'auth': auth,
      'iterate': iterate
  }
  return API(configuration)


def API_DBM(auth, iterate=False):
  """DBM helper configuration for Google API.

  Defines agreed upon version.
  """

  configuration = {
      'api': 'doubleclickbidmanager',
      'version': 'v1.1',
      'auth': auth,
      'iterate': iterate
  }
  return API(configuration)


def API_Sheets(auth, iterate=False):
  """DBM helper configuration for Google API.

  Defines agreed upon version.
  """

  configuration = {
      'api': 'sheets',
      'version': 'v4',
      'auth': auth,
      'iterate': iterate
  }
  return API(configuration)


def API_DCM(auth, iterate=False, internal=False):
  """DCM helper configuration for Google API.

  Defines agreed upon version.
  """

  configuration = {
      'api': 'dfareporting',
      'version': 'v3.4',
      'auth': auth,
      'iterate': iterate
  }

  if internal:
    from starthinker.util.dcm.internalv33_uri import URI as DCM_URI
    configuration['version'] = 'internalv3.3'
    configuration['uri'] = DCM_URI

  return API(configuration)


def API_SNIPPETS(auth, iterate=False):
  """Snippets helper configuration for Google API.

  Defines agreed upon version.
  """

  from starthinker.util.snippets.snippets_v1 import URI as SNIPPETS_URI

  # fetch discovery uri using: wget https://snippets-hrdb.googleplex.com/_ah/api/discovery/v1/apis/snippets/v1/rest
  configuration = {
      'api': 'snippets',
      'version': 'v1',
      'auth': auth,
      'iterate': iterate,
      'uri': SNIPPETS_URI
  }

  return API(configuration)


def API_Datastore(auth, iterate=False):
  """Datastore helper configuration for Google API.

  Defines agreed upon version.
  """

  configuration = {
      'api': 'datastore',
      'version': 'v1',
      'auth': auth,
      'iterate': iterate
  }
  return API(configuration)


def API_StackDriver(auth, iterate=False):
  """StackDriver helper configuration for Google API.

  Defines agreed upon version.
  """

  configuration = {
      'api': 'logging',
      'version': 'v2',
      'auth': auth,
      'iterate': iterate
  }
  return API(configuration)


def API_PubSub(auth, iterate=False):
  """PubSub helper configuration for Google API.

  Defines agreed upon version.
  """

  configuration = {
      'api': 'pubsub',
      'version': 'v1',
      'auth': auth,
      'iterate': iterate
  }
  return API(configuration)


def API_Analytics(auth, iterate=False):
  """Analytics helper configuration Google API.

  Defines agreed upon version.
  """

  configuration = {
      'api': 'analytics',
      'version': 'v3',
      'auth': auth,
      'iterate': iterate
  }
  return API(configuration)


def API_YouTube(auth, iterate=False):
  """YouTube helper configuration Google API.

  Defines agreed upon version.
  """

  configuration = {
      'api': 'youtube',
      'version': 'v3',
      'auth': auth,
      'iterate': iterate
  }
  return API(configuration)


def API_Drive(auth, iterate=False):
  """Drive helper configuration Google API.

  Defines agreed upon version.
  """

  configuration = {
      'api': 'drive',
      'version': 'v3',
      'auth': auth,
      'iterate': iterate
  }
  return API(configuration)


def API_Cloud(auth, iterate=False):
  """Cloud project helper configuration Google API.

  Defines agreed upon version.
  """

  configuration = {
      'api': 'cloudresourcemanager',
      'version': 'v1',
      'auth': auth,
      'iterate': iterate
  }
  return API(configuration)


def API_DV360(auth, iterate=False):
  """Cloud project helper configuration Google API.

  Defines agreed upon version.
  """

  configuration = {
      'api': 'displayvideo',
      'version': 'v1',
      'auth': auth,
      'iterate': iterate
  }
  return API(configuration)


def API_Storage(auth, iterate=False):
  """Cloud storage helper configuration Google API.

  Defines agreed upon version.
  """

  configuration = {
      'api': 'storage',
      'version': 'v1',
      'auth': auth,
      'iterate': iterate
  }
  return API(configuration)


def API_Gmail(auth, iterate=False):
  """Gmail helper configuration Google API.

  Defines agreed upon version.
  """

  configuration = {
      'api': 'gmail',
      'version': 'v1',
      'auth': auth,
      'iterate': iterate
  }
  return API(configuration)


def API_Compute(auth, iterate=False):
  """Compute helper configuration Google API.

     https://cloud.google.com/compute/docs/reference/rest/v1/
  """

  configuration = {
      'api': 'compute',
      'version': 'v1',
      'auth': auth,
      'iterate': iterate
  }
  return API(configuration)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import re
import json

from django import forms

from starthinker.script.parse import json_get_fields
from starthinker_ui.recipe.forms_fields import CommaSeparatedCharField, CommaSeparatedIntegerField, JsonField, TimezoneField, SwitchField


def load_config_data(config, path, default):
  current = config
  if isinstance(path, str):
    path = path.split('.')  # accept string or list as path
  for key in path:
    if not isinstance(key, int) and key.isdigit():
      key = int(key)  # cast string indexes to int
    try:
      current = current[key]
    except:
      return default
  return current


def get_field_kind(field):
  if field['kind'] == 'string':
    return forms.CharField(max_length=255, required=False)
  elif field['kind'] == 'email':
    return forms.EmailField(max_length=255, required=False)
  elif field['kind'] == 'integer':
    return forms.IntegerField(required=False)
  elif field['kind'] == 'boolean':
    return forms.BooleanField(required=False)
  elif field['kind'] == 'text':
    return forms.CharField(widget=forms.Textarea(), required=False)
  elif field['kind'] == 'choice':
    return forms.ChoiceField(choices=map(lambda c: (c, c), field['choices']))
  elif field['kind'] == 'timezones':
    return TimezoneField()
  elif field['kind'] == 'authentication':
    return SwitchField('user', 'service', required=True)
  elif field['kind'] == 'json':
    return JsonField(required=False)
  elif field['kind'] == 'integer_list':
    return CommaSeparatedIntegerField(required=False)
  elif field['kind'] == 'string_list':
    return CommaSeparatedCharField(required=False)
  else:
    return forms.CharField(max_length=255, required=False)


class ScriptJsonForm(forms.Form):
  script_sequence = forms.IntegerField(
      widget=forms.HiddenInput(attrs={'class': 'form_sequence'}))
  script_delete = forms.IntegerField(
      widget=forms.HiddenInput(attrs={'class': 'form_delete'}), initial=0)

  def __init__(self, sequence, script, values, *args, **kwargs):
    self.script = script
    self.values = values
    super(ScriptJsonForm, self).__init__(*args, **kwargs)
    self.fields['script_sequence'].initial = sequence
    x = script.get_tag()
    self.variables = json_get_fields(script.get_tasks())
    for variable in self.variables:
      if variable['name'].startswith('recipe_'):
        continue  # skip inputs that come from recipe constants
      self.fields[variable['name']] = get_field_kind(variable)
      self.fields[variable['name']].initial = values.get(
          variable['name'], variable.get('value', variable.get('default', '')))
      self.fields[variable['name']].required = variable.get('required', False)
      self.fields[variable['name']].help_text = variable.get('description', '')

  def get_description(self):
    return self.script.get_description(self.values)

  def get_instructions(self):
    return self.script.get_instructions(self.values)

  def get_script(self):
    # if marked for delete return nothing, will be filtered up stream
    if self.cleaned_data.get('script_delete', 0) == 1:
      return None
    # if saving, return values ( constants are not stored as they come from the recipe )
    else:
      values = {}
      for variable in self.variables:
        if variable['name'].startswith('recipe_'):
          continue  # skip inputs that come from recipe constants
        values[variable['name']] = self.cleaned_data[variable['name']]
      return {
          'tag': self.script.get_tag(),
          'values': values,
          'sequence': self.cleaned_data['script_sequence']
      }



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""This is a technical settings file that needs to be extended with additional settings.

   Django Specific Documentation:
     - Generated by 'django-admin startproject' using Django 1.11.1.
     - For more information on this file, see:
     https://docs.djangoproject.com/en/1.11/topics/settings/
     - For the full list of settings and their values, see:
     https://docs.djangoproject.com/en/1.11/ref/settings/
     - For production, see:
     https://docs.djangoproject.com/en/1.11/howto/deployment/checklist/

"""

import os

ADMIN_ENABLED = False

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'starthinker_ui.account.apps.AccountConfig',
    'starthinker_ui.recipe',
    'starthinker_ui.project',
    'starthinker_ui.website',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

# Relative to current application load
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [BASE_DIR + '/templates/',],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME':
            'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME':
            'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME':
            'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME':
            'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]

AUTHENTICATION_BACKENDS = [
    'django.contrib.auth.backends.ModelBackend',
]

USE_I18N = True
USE_L10N = True
USE_TZ = False

DEFAULT_CHARSET = 'utf-8'

# User
AUTH_USER_MODEL = 'account.Account'

# Application Handler Maps
ROOT_URLCONF = 'starthinker_ui.ui.urls'
WSGI_APPLICATION = 'starthinker_ui.ui.wsgi.application'



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from googleapiclient.errors import HttpError

from starthinker.util.bigquery import table_list
from starthinker.util.data import get_rows
from starthinker.util.google_api.discovery_to_bigquery import Discovery_To_BigQuery
from starthinker.util.project import project

from starthinker.task.google_api.run import google_api_build_errors
from starthinker.task.google_api.run import google_api_build_results
from starthinker.task.google_api.run import google_api_execute
from starthinker.task.google_api.run import google_api_initilaize


def build_request(endpoint):
  return {
    "bigquery": {
      "dataset": endpoint['dataset'],
      "table": endpoint['table']
    }
  }


def build_results(auth, api_call, endpoint):
  return google_api_build_results(
      auth, api_call, {
          'bigquery': {
              'dataset':
                  endpoint['dataset'],
              'table':
                  endpoint['table'].replace('BQFlow__', 'BQFlow__RESULTS__')
          }
      })


def build_errors(auth, api_call, endpoint):
  return google_api_build_errors(
      auth, api_call, {
          'bigquery': {
              'dataset': endpoint['dataset'],
              'table': endpoint['table'].replace('BQFlow__', 'BQFlow__ERRORS__')
          }
      })


@project.from_parameters
def bqflow():

  if project.verbose: print('BQFLOW')

  endpoints = []

  # load dataset / table list
  for dataset, table, kind in table_list(project.task['auth'], project.id):
    if table.startswith('BQFlow__') and not table.startswith('BQFlow__RESULTS__') and not table.startswith('BQFlow__ERRORS__'):
      print(table, kind)
      endpoints.append({'dataset': dataset, kind.lower(): table})

  for endpoint in endpoints:
    if 'table' in endpoint:
      _, api, function = endpoint['table'].split('__', 2)
      function = function.replace('__', '.')

      api_call = {
          'auth':
              'user',
          'api':
              api,
          'version':
              Discovery_To_BigQuery.preferred_version(api,
                                                      project.task.get('key')),
          'function':
              function,
      }

      kwargs_list = get_rows(
          project.task['auth'], build_request(endpoint), as_object=True)

      results = build_results(project.task['auth'], api_call, endpoint)
      errors = build_errors(project.task['auth'], api_call, endpoint)

      for kwargs in kwargs_list:
        api_call['kwargs'] = kwargs

        if project.verbose: print('BQFLOW API CALL:', api_call)

        google_api_initilaize(api_call)
        google_api_execute(project.task['auth'], api_call, results, errors)


if __name__ == '__main__':
  bqflow()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import argparse
import textwrap

from starthinker.util.project import project
from starthinker.util.csv import excel_to_sheets
from starthinker.util.csv import excel_to_rows
from starthinker.util.csv import rows_to_csv


def main():

  parser = argparse.ArgumentParser(
      formatter_class=argparse.RawDescriptionHelpFormatter,
      description=textwrap.dedent("""\
    Command line to transform excel sheets into csv files.

    Prints to STDOUT, user is expected to pipe output into file.
    Typically used for BigQuery data imports.

    Examples:
      List sheets in workbook: python helper.py [EXCEL FILE] --list
      Convert excel to CSV: python helper.py [EXCEL FILE] --sheet [SHEET NAME] > results.csv

  """))

  parser.add_argument('workbook', help='name of file to pull the rows.')
  parser.add_argument('--sheet', help='Sheet to pull the rows.', default=None)
  parser.add_argument('--list', help='List reports.', action='store_true')

  # initialize project ( not really used but consistent with all helpers )
  project.from_commandline(parser=parser, arguments=('-v',))

  with open(project.args.workbook, 'rb') as excel_file:
    if project.args.list:
      for sheet in excel_to_sheets(excel_file):
        print(sheet)
    elif project.args.sheet:
      for sheet, row in excel_to_rows(excel_file, project.args.sheet):
        print(rows_to_csv(row).read())


if __name__ == '__main__':
  main()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer from open source:

    pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory:

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

CM Standard Bulk

Aggregate multiple standard CM reports into one BigQuery or Sheet.

See API docs for <a
href='https://developers.google.com/doubleclick-advertisers/v3.2/dimensions'
target='_blank'>Metrics</a>.
CM report name format '[Report Name] [Account ID] ( StarThinker )'.
Specify either bucket and path or dataset and table.
Schema is pulled from the official CM specification.

"""

from starthinker_airflow.factory import DAG_Factory

# Add the following credentials to your Airflow configuration.
USER_CONN_ID = 'starthinker_user'  # The connection to use for user authentication.
GCP_CONN_ID = 'starthinker_service'  # The connection to use for service authentication.

INPUTS = {
    'accounts': '',
    'name': '',
    'range': 'LAST_7_DAYS',
    'dcm_dimensions': [
        'date', 'platformType', 'creativeType', 'state', 'dmaRegion'
    ],
    'dcm_metrics': ['impressions'],
    'dataset': '',
    'table': '',
    'bucket': '',
    'path': 'CM_Report',
    'delete': False,
    'datastudio': True,
}

TASKS = [{
    'dcm_bulk': {
        'auth': 'user',
        'accounts': {
            'field': {
                'name': 'accounts',
                'kind': 'integer_list',
                'order': 1,
                'default': ''
            }
        },
        'name': {
            'field': {
                'name': 'name',
                'kind': 'string',
                'order': 2,
                'default': ''
            }
        },
        'report': {
            'type': 'STANDARD',
            'timeout': 0,
            'relativeDateRange': {
                'field': {
                    'name':
                        'range',
                    'kind':
                        'choice',
                    'order':
                        3,
                    'default':
                        'LAST_7_DAYS',
                    'choices': [
                        'LAST_24_MONTHS', 'LAST_30_DAYS', 'LAST_365_DAYS',
                        'LAST_7_DAYS', 'LAST_90_DAYS', 'MONTH_TO_DATE',
                        'PREVIOUS_MONTH', 'PREVIOUS_QUARTER', 'PREVIOUS_WEEK',
                        'PREVIOUS_YEAR', 'QUARTER_TO_DATE', 'TODAY',
                        'WEEK_TO_DATE', 'YEAR_TO_DATE', 'YESTERDAY'
                    ]
                }
            },
            'dimensions': {
                'field': {
                    'name':
                        'dcm_dimensions',
                    'kind':
                        'string_list',
                    'order':
                        4,
                    'default': [
                        'date', 'platformType', 'creativeType', 'state',
                        'dmaRegion'
                    ]
                }
            },
            'metrics': {
                'field': {
                    'name': 'dcm_metrics',
                    'kind': 'string_list',
                    'order': 5,
                    'default': ['impressions']
                }
            }
        },
        'out': {
            'bigquery': {
                'dataset': {
                    'field': {
                        'name': 'dataset',
                        'kind': 'string',
                        'order': 5,
                        'default': ''
                    }
                },
                'table': {
                    'field': {
                        'name': 'table',
                        'kind': 'string',
                        'order': 6,
                        'default': ''
                    }
                }
            },
            'storage': {
                'bucket': {
                    'field': {
                        'name': 'bucket',
                        'kind': 'string',
                        'order': 7,
                        'default': ''
                    }
                },
                'path': {
                    'field': {
                        'name': 'path',
                        'kind': 'string',
                        'order': 8,
                        'default': 'CM_Report'
                    }
                }
            }
        },
        'delete': {
            'field': {
                'name': 'delete',
                'kind': 'boolean',
                'order': 10,
                'default': False
            }
        },
        'datastudio': {
            'field': {
                'name': 'datastudio',
                'kind': 'boolean',
                'order': 11,
                'default': True
            }
        }
    }
}]

DAG_FACTORY = DAG_Factory('dcm_bulk_standard', {'tasks': TASKS}, INPUTS)
DAG_FACTORY.apply_credentails(USER_CONN_ID, GCP_CONN_ID)
DAG = DAG_FACTORY.execute()

if __name__ == '__main__':
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import argparse
import json
import pprint
import textwrap

from starthinker.util.google_api import API
from starthinker.util.google_api.discovery_to_bigquery import Discovery_To_BigQuery
from starthinker.util.project import project


def main():

  parser = argparse.ArgumentParser(
      formatter_class=argparse.RawDescriptionHelpFormatter,
      description=textwrap.dedent("""\
      Command line interface for running Google API calls.  Any API works.  Allows developers to quickly test
      and debug API calls before building them into scripts.  Useful for debugging permission or call errors.

      Examples:
        - Pull a DBM report via API.
          - https://developers.google.com/bid-manager/v1/queries/getquery
          - python google_api/helper.py -api doubleclickbidmanager -version v1 -function queries.getquery -kwargs '{ "queryId": 132865172 }' -u [credentials path]

        - Pull a list of placements:
          - https://developers.google.com/doubleclick-advertisers/v3.3/placements/list
          - python task/google_api/helper.py -api dfareporting -version v3.3 -function placements.list -kwargs '{ "profileId":2782211 }' -u [credentials path]

        - Show schema for Campaign Manager advertiser list endpoint.
        - https://developers.google.com/doubleclick-advertisers/v3.4/advertisers/list
        - python starthinker/task/google_api/helper.py -api dfareporting -version v3.4 -function advertisers.list --schema

  """))

  # get parameters
  parser.add_argument('-api', help='api to run, name of product api')
  parser.add_argument('-version', help='version of api')
  parser.add_argument('-function', help='function to call in api')
  parser.add_argument('-uri', help='uri to use in api', default=None)
  parser.add_argument(
      '-developer-token',
      help='developer token to pass in header',
      default=None)
  parser.add_argument(
      '-login-customer-id',
      help='customer to log in with when manipulating an MCC',
      default=None)
  parser.add_argument(
      '-kwargs',
      help='kwargs to pass to function, json string of name:value pairs')
  parser.add_argument('--iterate', help='force iteration', action='store_true')
  parser.add_argument(
      '--schema',
      help='return schema instead, function = [endpoint.method]',
      action='store_true')


  # initialize project ( used to load standard credentials parameters )
  project.from_commandline(parser=parser, arguments=('-u', '-c', '-s', '-k', '-v'))

  # show schema
  if project.args.schema:
    endpoint, method = project.args.function.rsplit('.', 1)
    print(json.dumps(Discovery_To_BigQuery(project.args.api, project.args.version).method_schema(endpoint, method), indent=2, default=str))

  # or fetch results
  else:

    # the api wrapper takes parameters as JSON
    job = {
      'auth': 'service' if project.args.service else 'user',
      'api': project.args.api,
      'version': project.args.version,
      'function': project.args.function,
      'key': project.args.key,
      'uri': project.args.uri,
      'kwargs': json.loads(project.args.kwargs),
      'headers': {},
      'iterate': project.args.iterate,
    }

    if project.args.developer_token:
      job['headers']['developer-token'] = project.args.developer_token

    if project.args.login_customer_id:
      job['headers']['login-customer-id'] = project.args.login_customer_id

    # run the API call
    results = API(job).execute()

    # display results
    if project.args.iterate:
      for result in results:
        pprint.PrettyPrinter().pprint(result)
    else:
      pprint.PrettyPrinter().pprint(results)


if __name__ == '__main__':
  main()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import sys
import socket
import threading

from googleapiclient import discovery
from googleapiclient.http import HttpRequest

from starthinker.util.auth.wrapper import CredentialsFlowWrapper
from starthinker.util.auth.wrapper import CredentialsServiceWrapper
from starthinker.util.auth.wrapper import CredentialsUserWrapper
from starthinker.util.project import project

# WARNING:  possible issue if switching user credentials mid recipe, not in scope but possible ( need to address using hash? )
CREDENTIALS_USER_CACHE = None
DISCOVERY_CACHE = {}

# set timeout to 10 minutes ( reduce socket.timeout: The read operation timed out )
socket.setdefaulttimeout(600)

def clear_credentials_cache():
  global CREDENTIALS_USER_CACHE
  CREDENTIALS_USER_CACHE = None


def get_credentials(auth):
  global CREDENTIALS_USER_CACHE

  if auth == 'user':
    if CREDENTIALS_USER_CACHE is None:
      try:
        CREDENTIALS_USER_CACHE = CredentialsUserWrapper(
          project.recipe['setup']['auth']['user'],
          project.recipe['setup']['auth'].get('client')
        )
      except (KeyError, ValueError):
        print('')
        print(
            'ERROR: You are attempting to access an API endpoiont that requires Google OAuth USER authentication but have not provided credentials to make that possible.'
        )
        print('')
        print(
            'SOLUTION: Specify a -u [user credentials path] parameter on the command line.'
        )
        print(
            '          Alternaitvely specify a -u [user credentials path to be created] parameter and a -c [client credentials path] parameter on the command line.'
        )
        print(
            '          Alternaitvely if running a recipe, include { "setup":{ "auth":{ "user":"[JSON OR PATH]" }}} in the JSON.'
        )
        print('')
        print(
            'INSTRUCTIONS: https://github.com/google/starthinker/blob/master/tutorials/cloud_client_installed.md'
        )
        print('')
        sys.exit(1)

    return CREDENTIALS_USER_CACHE

  elif auth == 'service':
    try:
      return CredentialsServiceWrapper(
        project.recipe['setup']['auth']['service']
      )
    except (KeyError, ValueError):
      print('')
      print(
          'ERROR: You are attempting to access an API endpoint that requires Google Cloud SERVICE authentication but have not provided credentials to make that possible.'
      )
      print('')
      print(
          'SOLUTION: Specify a -s [service credentials path] parameter on the command line.'
      )
      print(
          '          Alternaitvely if running a recipe, include { "setup":{ "auth":{ "service":"[JSON OR PATH]" }}} in the JSON.'
      )
      print('')
      print(
          'INSTRUCTIONS: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md'
      )
      print('')
      sys.exit(1)


def get_service(api='gmail',
                version='v1',
                auth='service',
                scopes=None,
                headers=None,
                key=None,
                uri_file=None):
  global DISCOVERY_CACHE

  class HttpRequestCustom(HttpRequest):

    def __init__(self, *args, **kwargs):
      if headers:
        kwargs['headers'].update(headers)
      super(HttpRequestCustom, self).__init__(*args, **kwargs)

  if not key:
    key = project.recipe['setup'].get(key, '')

  cache_key = api + version + auth + key + str(threading.current_thread().ident)

  if cache_key not in DISCOVERY_CACHE:
    credentials = get_credentials(auth)
    if uri_file:
      uri_file = uri_file.strip()
      if uri_file.startswith('{'):
        DISCOVERY_CACHE[cache_key] = discovery.build_from_document(
            uri_file,
            credentials=credentials,
            developerKey=key,
            requestBuilder=HttpRequestCustom)
      else:
        with open(uri_file, 'r') as cache_file:
          DISCOVERY_CACHE[cache_key] = discovery.build_from_document(
              cache_file.read(),
              credentials=credentials,
              developerKey=key,
              requestBuilder=HttpRequestCustom)
    else:
      DISCOVERY_CACHE[cache_key] = discovery.build(
          api,
          version,
          credentials=credentials,
          developerKey=key,
          requestBuilder=HttpRequestCustom)

  return DISCOVERY_CACHE[cache_key]


def get_client_type(credentials):
  client_json = CredentialsFlowWrapper(credentials, credentials_only=True)
  return next(iter(client_json.keys()))


def get_profile():
  service = get_service('oauth2', 'v2', 'user')
  return service.userinfo().get().execute()


def set_iam(auth, project_id, role, email=None, service=None):
  service = get_service('cloudresourcemanager', 'v1', auth,
                        ['https://www.googleapis.com/auth/cloud-platform'])
  policy = service.projects().getIamPolicy(
      resource=project_id, body={}).execute()

  if email:
    member = 'user:%s' % email
  elif service:
    member = 'service:%s' % service
  else:
    member = None

  if member:
    for r in policy['bindings']:
      if r['role'] == role and member not in r['members']:
        if project.verbose:
          print('Settings Role:', member, role)
        r['members'].append(member)
        policy = service.projects().setIamPolicy(
            resource=project_id, body={
                'policy': policy
            }).execute()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from googleapiclient.errors import HttpError

from starthinker.util.bigquery import table_create
from starthinker.util.data import get_rows
from starthinker.util.data import put_rows
from starthinker.util.dcm import get_profile_for_api
from starthinker.util.google_api import API
from starthinker.util.google_api.discovery_to_bigquery import Discovery_To_BigQuery
from starthinker.util.project import project


ERROR_SCHEMA = [
  { 'name': 'Error', 'type': 'STRING', 'mode': 'NULLABLE' },
  { 'name': 'Parameters', 'type': 'RECORD', 'mode': 'REPEATED', 'fields': [
    { 'name': 'Key', 'type': 'STRING', 'mode': 'NULLABLE' },
    { 'name': 'Value', 'type': 'STRING', 'mode': 'NULLABLE' },
  ]}
]


def google_api_initilaize(api_call, alias=None):

  if api_call['function'].endswith('list') or alias == 'list':
    api_call['iterate'] = True

  if api_call['api'] == 'dfareporting':
    is_superuser, profile_id = get_profile_for_api(
        api_call['auth'], api_call['kwargs']['accountId'])

    api_call['kwargs']['profileId'] = profile_id

    if is_superuser:
      from starthinker.util.dcm.internalv33_uri import URI as DCM_URI
      api_call['version'] = 'internalv3.3'
      api_call['uri'] = DCM_URI
    else:
      del api_call['kwargs']['accountId']


def google_api_build_results(auth, api_call, results):
  if 'bigquery' in results:
    results['bigquery']['schema'] = Discovery_To_BigQuery(
        api_call['api'],
        api_call['version'],
        api_call.get('key', None),
    ).method_schema(api_call['function'])

    #TODO: Fix format to sometimes be CSV, probably refactor BigQuery to
    # determine format based on rows or schema
    results['bigquery']['format'] = 'JSON'
    results['bigquery']['skip_rows'] = 0
    results['bigquery']['disposition'] = 'WRITE_TRUNCATE'

    table_create(
        results['bigquery'].get('auth', auth),
        project.id,
        results['bigquery']['dataset'],
        results['bigquery']['table'],
        results['bigquery']['schema'],
        overwrite=False)

  return results


def google_api_build_errors(auth, api_call, errors):
  if 'bigquery' in errors:
    errors['bigquery']['schema'] = ERROR_SCHEMA
    errors['bigquery']['format'] = 'JSON'
    errors['bigquery']['skip_rows'] = 0
    errors['bigquery']['disposition'] = 'WRITE_TRUNCATE'

    table_create(
        errors['bigquery'].get('auth', auth),
        project.id,
        errors['bigquery']['dataset'],
        errors['bigquery']['table'],
        errors['bigquery']['schema'],
        overwrite=False)

  return errors


def google_api_execute(auth, api_call, results, errors):
  try:
    rows = API(api_call).execute()

    if results:
      # check if single object needs conversion to rows
      if isinstance(rows, dict):
        rows = [rows]
      rows = map(lambda r: Discovery_To_BigQuery.clean(r), rows)
      put_rows(auth, results, rows)

      if 'bigquery' in results:
        results['bigquery']['disposition'] = 'WRITE_APPEND'

  except HttpError as e:

    if errors:
      rows = [{
          'Error':
              str(e),
          'Parameters': [{
              'Key': k,
              'Value': str(v)
          } for k, v in api_call['kwargs'].items()]
      }]
      put_rows(auth, errors, rows)

      if 'bigquery' in errors:
        errors['bigquery']['disposition'] = 'WRITE_APPEND'

    else:
      raise e


@project.from_parameters
def google_api():

  if project.verbose:
    print('GOOGLE_API', project.task['api'], project.task['version'],
          project.task['function'])

  api_call = {
      'auth': project.task['auth'],
      'api': project.task['api'],
      'version': project.task['version'],
      'function': project.task['function'],
      'iterate': project.task.get('iterate', False),
      'key': project.key,
      'headers': project.task.get('headers'),
  }

  results = google_api_build_results(
    project.task['auth'],
    api_call,
    project.task.get('results', {})
  )

  errors = google_api_build_errors(
    project.task['auth'],
    api_call,
    project.task.get('errors', {})
  )

  if 'kwargs' in project.task:
    kwargs_list = project.task['kwargs'] if isinstance(
        project.task['kwargs'], (list, tuple)) else [project.task['kwargs']]
  elif 'kwargs_remote' in project.task:
    kwargs_list = get_rows(
        project.task['auth'], project.task['kwargs_remote'], as_object=True)
  else:
    kwargs_list = [{}]

  for kwargs in kwargs_list:

    api_call['kwargs'] = kwargs

    google_api_initilaize(api_call, project.task.get('alias'))

    google_api_execute(project.task['auth'], api_call, results, errors)


if __name__ == '__main__':
  google_api()



import json

from starthinker.util.project import project
from starthinker.util.google_api import API_Drive
from starthinker.util.bigquery import query_to_rows


def get_drive_authorized_users(auth, drive, file_id):
  drive_file = drive.files().get(
      fileId=project.task['file_id'], fields='name,permissions').execute()
  result = {}

  for permission in drive_file['permissions']:
    if permission.get('emailAddress', None):
      result[permission['emailAddress']] = permission['id']

  return result


def get_dbm_authorization(auth, emails):
  result = {}

  in_email = '\'%s\'' % '\',\''.join(emails)
  query = ('SELECT email, primaryEmail, super_user, global_access, '
           'global_reporting_user, permissions FROM '
           '[plx.google:mauriciod.starthinker_dbmsentinel.all] WHERE email in '
           '(%s) or primaryEmail in (%s) ') % (in_email, in_email)

  access_list = query_to_rows(
      auth,
      'google.com:starthinker',
      'starthinker',
      query,
      row_max=None,
      legacy=True)

  for row in access_list:
    access = {
        'email': row[0],
        'primary_email': row[1] == 'true',
        'super_user': row[2] == 'true',
        'global_access': row[3] == 'true',
        'global_reporting_user': row[4],
        'permissions': row[5] or 'NOPERMISSIONS'
    }

    for email in emails:
      if email.lower() == access['email'].lower() or email.lower(
      ) == access['primary_email']:
        result[email] = access

  return result


def is_dbm_super_user(access):
  return access['super_user'] or access['global_access']


def has_dbm_partner_access(access, partner_ids):
  if not partner_ids:
    return False

  for partner_id in partner_ids:
    if access.get('permissions', '').find('P%s' % partner_id) == -1:
      return False

  return True


def has_dbm_advertiser_access(access, advertiser_ids):
  if not advertiser_ids:
    return False

  for advertiser_id in advertiser_ids:
    if access.get('permissions', '').find('A%s' % advertiser_id) == -1:
      return False

  return True


def get_dbm_unauthorized_users(emails, dbm_authorization, partner_ids,
                               advertiser_ids):
  result = []

  for email in emails:
    if email in dbm_authorization:
      access = dbm_authorization[email]

      if not is_dbm_super_user(access):
        if not has_dbm_partner_access(access, partner_ids):
          if not has_dbm_advertiser_access(access, advertiser_ids):
            result.append(email)
    else:
      # No user access found
      result.append(email)

  return result


@project.from_parameters
def sentinel():
  if project.verbose:
    print('Sentinel')

  auth = project.task['auth']

  partner_ids = None
  partner_ids_config = project.task.get('dbm_partner_ids', '')
  if partner_ids_config:
    partner_ids = [
        partner_id.strip() for partner_id in partner_ids_config.split(',')
    ]

  advertiser_ids = None
  advertiser_ids_config = project.task.get('dbm_advertiser_ids', '')
  if advertiser_ids_config:
    advertiser_ids = [
        advertiser_id.strip()
        for advertiser_id in advertiser_ids_config.split(',')
    ]

  authorized_users = get_drive_authorized_users(auth, drive,
                                                project.task['file_id'])

  dbm_authorization = get_dbm_authorization(auth, authorized_users.keys())
  unauthorized_users = get_dbm_unauthorized_users(authorized_users.keys(),
                                                  dbm_authorization,
                                                  partner_ids, advertiser_ids)

  for user in unauthorized_users:
    API_Drive(auth).permissions().delete(
        fileId=project.task['file_id'],
        permissionId=authorized_users[user]).execute()


if __name__ == '__main__':
  sentinel()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

Standard_Report_Metrics_Schema = [{
    'name': 'Active_View_Average_Viewable_Time_Seconds',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Eligible_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name':
        'Active_View_Audible_Amp_Fully_On_Screen_For_Half_Of_Duration_15_Sec_Cap_Measurable_Impressions',
    'type':
        'INTEGER',
    'mode':
        'NULLABLE'
}, {
    'name':
        'Active_View_Audible_Amp_Fully_On_Screen_For_Half_Of_Duration_Trueview_Measurable_Impressions',
    'type':
        'INTEGER',
    'mode':
        'NULLABLE'
}, {
    'name':
        'Active_View_Audible_Amp_Fully_On_Screen_For_Half_Of_Duration_Trueview_Impressions',
    'type':
        'INTEGER',
    'mode':
        'NULLABLE'
}, {
    'name':
        'Active_View_Audible_Amp_Fully_On_Screen_For_Half_Of_Duration_15_Sec_Cap_Impressions',
    'type':
        'INTEGER',
    'mode':
        'NULLABLE'
}, {
    'name': 'Active_View_Impression_Distribution_Not_Measurable',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Impression_Distribution_Not_Viewable',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Impression_Distribution_Viewable',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Impressions_Audible_And_Visible_At_Completion',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Impressions_Visible_10_Seconds',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Measurable_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Not_Measurable_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Not_Viewable_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_Audible_And_Visible_At_Completion',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_Audible_And_Visible_At_First_Quartile',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_Audible_And_Visible_At_Midpoint',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_Audible_And_Visible_At_Start',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_Audible_And_Visible_At_Third_Quartile',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_Audible_Impressions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_In_Background',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_Full_Screen',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name':
        'Active_View_Audible_Amp_Fully_On_Screen_For_Half_Of_Duration_15_Sec_Cap_Rate',
    'type':
        'FLOAT',
    'mode':
        'NULLABLE'
}, {
    'name':
        'Active_View_Audible_Amp_Fully_On_Screen_For_Half_Of_Duration_Trueview_Rate',
    'type':
        'FLOAT',
    'mode':
        'NULLABLE'
}, {
    'name': 'Active_View_Percent_Of_Completed_Impressions_Audible_And_Visible',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name':
        'Active_View_Percent_Of_First_Quartile_Impressions_Audible_And_Visible',
    'type':
        'FLOAT',
    'mode':
        'NULLABLE'
}, {
    'name': 'Active_View_Percent_Of_Midpoint_Impressions_Audible_And_Visible',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name':
        'Active_View_Percent_Of_Third_Quartile_Impressions_Audible_And_Visible',
    'type':
        'FLOAT',
    'mode':
        'NULLABLE'
}, {
    'name': 'Active_View_Percent_Of_Completed_Impressions_Visible',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_Of_First_Quartile_Impressions_Visible',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_Of_Midpoint_Impressions_Visible',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_Of_Third_Quartile_Impressions_Visible',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_Play_Time_Audible',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_Play_Time_Audible_And_Visible',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_Play_Time_Visible',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_Visible_10_Seconds',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_Visible_At_Completion',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_Visible_At_First_Quartile',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_Visible_At_Midpoint',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_Visible_At_Start',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_Visible_At_Third_Quartile',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_Measurable_Impressions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Percent_Viewable_Impressions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Active_View_Viewable_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Click_Through_Conversions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Click_Through_Revenue',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Activity_Delivery_Status',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Activity_Per_Click',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Activity_Per_Thousand_Impressions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'View_Through_Conversions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'View_Through_Revenue',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Audio_Companion_Clicks',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Audio_Companion_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Booked_Activities',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Booked_Clicks',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Booked_Impressions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Booked_Viewable_Impressions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Click_Delivery_Status',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Click_Rate',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Clicks',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Cost_Per_Activity',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Cost_Per_Click',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Cost_Per_Revenue',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Click_Through_Conversions_Cross_Environment',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Click_Through_Revenue_Cross_Environment',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Click_Through_Conversion_Events_Cross_Environment',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Conversions_Cross_Environment',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Revenue_Cross_Environment',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Conversion_Events_Cross_Environment',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'View_Through_Conversions_Cross_Environment',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'View_Through_Revenue_Cross_Environment',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'View_Through_Conversion_Events_Cross_Environment',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Cost_Account_Currency',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Dbm_Cost_Usd',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Downloaded_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_Click_Rate',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_Click_Through_Conversions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_Clicks',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_Total_Conversions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_View_Through_Conversions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Effective_Cpm',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'General_Invalid_Traffic_Givt_Clicks',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'General_Invalid_Traffic_Givt_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Impression_Delivery_Status',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Invalid_Clicks',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Invalid_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Media_Cost',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Actions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Clicks',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Revenue',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Transactions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Paid_Search_Action_Conversion_Percentage',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Actions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Paid_Search_Average_Cost_Per_Action',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Paid_Search_Average_Cost_Per_Transaction',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Paid_Search_Average_Dcm_Transaction_Amount',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Average_Position',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Click_Rate',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Clicks',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Cost',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Revenue',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Paid_Search_Spend_Per_Transaction_Revenue',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Paid_Search_Transaction_Conversion_Percentage',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Paid_Search_Transaction_Revenue_Per_Spend',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Transactions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Visits',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Percent_Invalid_Clicks',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Percent_Invalid_Impressions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Planned_Media_Cost',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Revenue_Per_Click',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Revenue_Per_Thousand_Impressions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Audio_Completions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Audio_First_Quartile_Completions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Audio_Midpoints',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Audio_Mutes',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Audio_Pauses',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Audio_Plays',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Audio_Replays',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Audio_Third_Quartile_Completions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Audio_Unmutes',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Average_Display_Time',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Average_Expansion_Time',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Full_Screen_Average_View_Time',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Average_Interaction_Time',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Average_View_Time',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Backup_Image',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Rich_Media_Click_Rate',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Rich_Media_Clicks',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Code_Serves',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Average_Time',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Counters',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Exits',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Timers',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Custom_Variable_Count_1',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Custom_Variable_Count_2',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Display_Time',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Event_Counters',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Event_Timers',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Expansion_Time',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Expansions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Full_Screen_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Full_Screen_Video_Completions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Full_Screen_Video_Plays',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Html5_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Rich_Media_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Rate',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Interaction_Time',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Interactions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interactive_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Manual_Closes',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Trueview_Views',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Companion_Clicks',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Completions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_First_Quartile_Completions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Full_Screen',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Interaction_Rate',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Interactions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Midpoints',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Mutes',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Pauses',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Plays',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Progress_Events',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Replays',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Skips',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Stops',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Third_Quartile_Completions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Unmutes',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_View_Rate',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Views',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Roadblock_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Email_Shares',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Social_Engagement_Rate',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Follows',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Twitter_Leads_Generated',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Likes',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Twitter_Offers_Accepted',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Replies',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Social_Engagements',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Conversions',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Total_Revenue',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Twitter_App_Install_Clicks',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Twitter_Buy_Now_Clicks',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Twitter_Buy_Now_Purchases',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Other_Twitter_Engagements',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Retweets',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Twitter_Url_Clicks',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Twitter_Video_50Percent_In_View_For_2_Seconds',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Twitter_Video_100Percent_In_View_For_3_Seconds',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Blocked_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Classified_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Measurable_Impressions_For_Audio',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Measurable_Impressions_For_Video_Player_Location',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Measurable_Impressions_For_Video_Player_Size',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Percentage_Of_Measurable_Impressions_For_Audio',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Percentage_Of_Measurable_Impressions_For_Video_Player_Location',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Percentage_Of_Measurable_Impressions_For_Video_Player_Size',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Publisher_Problems',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Reporting_Problems',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Serving_Problems',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Verifiable_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Muted_At_Start',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Hd_Video_Player_Size_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Large_Video_Player_Size_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Player_Location_Avg_Pixels_From_Left',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Player_Location_Avg_Pixels_From_Top',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Player_Size_Avg_Height',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Player_Size_Avg_Width',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Small_Video_Player_Size_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Prominence_Score',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}, {
    'name': 'Warnings',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Companion_Clicks',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Companion_Impressions',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import json
import copy
import pytz
from time import time, sleep
from datetime import date, datetime, timedelta

from django.core import management
from django.conf import settings
from django.test import TestCase
from django.test.testcases import TransactionTestCase

from starthinker_ui.account.tests import account_create
from starthinker_ui.project.tests import project_create
from starthinker_ui.recipe.models import Recipe, utc_milliseconds, utc_to_timezone, timezone_to_utc, utc_milliseconds_to_timezone, JOB_INTERVAL_MS, JOB_LOOKBACK_MS, time_ago
from starthinker_ui.recipe.views import autoscale
from starthinker_ui.recipe.management.commands.job_worker import Workers, worker_pull, worker_status, worker_ping, worker_downscale

# test recipe done to undone

# starthinker/gtech/script_test.json
#
# Hours will equal 1 + 1 + 1 + n + 3 + 1 where n is the number of hours given in the setup
#
#  { "hello":{ +1 hour
#      "auth":"user",
#      "hour":[1],
#      "say":"Hello At 1",
#      "sleep":0
#    }},
#    { "hello":{ +1 hour
#      "auth":"user",
#      "hour":[3],
#      "say":"Hello At 3",
#      "sleep":0
#    }},
#    { "hello":{ +1 hour
#      "auth":"user",
#      "hour":[23],
#      "say":"Hello At 23 Sleep",
#      "sleep":30
#    }},
#    { "hello":{ + n hours
#      "auth":"user",
#      "say":"Hello At 1, 3, 23 Default",
#      "sleep":0
#    }},
#    { "hello":{ + 3 hours
#      "auth":"user",
#      "hour":[1, 3, 23],
#      "say":"Hello At 1, 3, 23 Explicit",
#      "sleep":0
#    }},
#    { "hello":{ + 1 hour
#      "auth":"user",
#      "hour":[3],
#      "say":"Hello At 3 Reorder",
#      "sleep":0
#    }}

WORKER_LOOKBACK_EXPIRE = (JOB_LOOKBACK_MS + 1000)


def test_job_utm(hour=0, days_offset=0):
  utc = datetime.utcnow().replace(hour=hour, minute=0, second=0, microsecond=0)
  if days_offset:
    utc += timedelta(days=days_offset)
  return utc_milliseconds(utc)


def assertRecipeDone(cls, recipe):
  recipe.refresh_from_db()
  status = recipe.get_status()
  job_time = datetime.utcfromtimestamp(int(recipe.job_utm / 1000))

  cls.assertTrue(all(task['done'] == True for task in status['tasks']))
  if recipe.manual:
    cls.assertEqual(recipe.job_utm, 0)
  else:
    cls.assertGreater(recipe.job_utm, utc_milliseconds())
  cls.assertEqual(job_time.minute, 0)


def assertRecipeNotDone(cls, recipe):
  recipe.refresh_from_db()
  status = recipe.get_status()
  job_time = datetime.utcfromtimestamp(int(recipe.job_utm / 1000))

  cls.assertFalse(all(task['done'] == True for task in status['tasks']))
  cls.assertLessEqual(recipe.job_utm, utc_milliseconds())
  cls.assertEqual(job_time.minute, 0)


class StatusTest(TestCase):

  def setUp(self):
    self.account = account_create()
    self.project = project_create()

    self.recipe = Recipe.objects.create(
        account=self.account,
        project=self.project,
        name='RECIPE_NEW',
        active=True,
        week=json.dumps(['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']),
        hour=json.dumps([1, 3, 23]),
        timezone='America/Los_Angeles',
        tasks=json.dumps([
            {
                'tag': 'test',
                'values': {},
                'sequence': 1
            },
        ]),
    )

  def test_interval(self):
    self.assertNotEqual(JOB_INTERVAL_MS / 1000, 0)

  def test_time(self):
    utc_now = datetime.utcnow().replace(second=0, microsecond=0)
    tz_now = datetime.now(tz=pytz.timezone(self.recipe.timezone)).replace(
        second=0, microsecond=0)

    utm = utc_milliseconds(utc_now)
    tz1 = utc_milliseconds_to_timezone(utm, self.recipe.timezone)
    tz2 = utc_to_timezone(utc_now, self.recipe.timezone)
    tz3 = timezone_to_utc(tz_now)

    self.assertEqual(tz1, tz2)
    self.assertEqual(tz3, utc_now)

  def test_schedule(self):
    tz_now = datetime.now(tz=pytz.timezone(self.recipe.timezone)).replace(
        second=0, microsecond=0)

    # pull new recipe, nothing in it until update is called
    status = self.recipe.get_status()

    self.assertEqual(status['date_tz'], str(tz_now.date()))
    self.assertEqual(status['tasks'], [])

    # causes recipe status to be filled in
    status = self.recipe.update()

    # new recipe
    job_utm = self.recipe.get_job_utm(status)
    tz_recipe = utc_milliseconds_to_timezone(job_utm, self.recipe.timezone)

    self.assertEqual(tz_recipe.date(), tz_now.date() + timedelta(days=1))
    self.assertEqual(tz_recipe.hour, self.recipe.get_hours()[0])

    # done recipe
    for task in status['tasks']:
      task['done'] = True

    job_utm = self.recipe.get_job_utm(status)
    tz_recipe = utc_milliseconds_to_timezone(job_utm, self.recipe.timezone)

    self.assertEqual(tz_recipe.date(), tz_now.date() + timedelta(days=1))
    self.assertEqual(tz_recipe.hour, self.recipe.get_hours()[0])

    # forced recipe
    status = self.recipe.force()

    job_utm = self.recipe.get_job_utm(status)
    tz_recipe = utc_milliseconds_to_timezone(job_utm, self.recipe.timezone)

    self.assertEqual(tz_recipe.date(), tz_now.date())
    self.assertEqual(tz_recipe.hour, tz_now.hour)

    # cancelled recipe
    status = self.recipe.cancel()

    job_utm = self.recipe.get_job_utm(status)
    tz_recipe = utc_milliseconds_to_timezone(job_utm, self.recipe.timezone)

    self.assertEqual(tz_recipe.date(), tz_now.date() + timedelta(days=1))
    self.assertEqual(tz_recipe.hour, self.recipe.get_hours()[0])

  def test_status(self):
    status = self.recipe.update()

    now_tz = utc_to_timezone(datetime.utcnow(), self.recipe.timezone)
    date_tz = str(now_tz.date())
    hour_tz = now_tz.hour

    self.assertEqual(status['date_tz'], date_tz)
    self.assertEqual(len(status['tasks']), 1 + 1 + 1 + 3 + 3 + 1)

    # order two is hidden unless forced ( so it is skipped ) hour = [] excludes it from list
    self.assertEqual(status['tasks'][0]['hour'], 1)
    self.assertEqual(status['tasks'][0]['order'], 0)
    self.assertEqual(status['tasks'][1]['hour'], 1)
    self.assertEqual(status['tasks'][1]['order'], 4)
    self.assertEqual(status['tasks'][2]['hour'], 1)
    self.assertEqual(status['tasks'][2]['order'], 5)
    self.assertEqual(status['tasks'][3]['hour'], 3)
    self.assertEqual(status['tasks'][3]['order'], 1)
    self.assertEqual(status['tasks'][4]['hour'], 3)
    self.assertEqual(status['tasks'][4]['order'], 4)
    self.assertEqual(status['tasks'][5]['hour'], 3)
    self.assertEqual(status['tasks'][5]['order'], 5)
    self.assertEqual(status['tasks'][6]['hour'], 3)
    self.assertEqual(status['tasks'][6]['order'], 6)
    self.assertEqual(status['tasks'][7]['hour'], 23)
    self.assertEqual(status['tasks'][7]['order'], 3)
    self.assertEqual(status['tasks'][8]['hour'], 23)
    self.assertEqual(status['tasks'][8]['order'], 4)
    self.assertEqual(status['tasks'][9]['hour'], 23)
    self.assertEqual(status['tasks'][9]['order'], 5)

  def test_all_hours(self):

    # reduce the hours so that fewer permutations are created
    self.recipe.hour = json.dumps([])
    self.recipe.save()
    self.recipe.update()

    status = self.recipe.get_status()

    now_tz = utc_to_timezone(datetime.utcnow(), self.recipe.timezone)
    date_tz = str(now_tz.date())

    self.assertEqual(status['date_tz'], date_tz)
    self.assertEqual(len(status['tasks']), 1 + 1 + 1 + 0 + 3 + 1)

  def test_forced(self):
    status = self.recipe.force()

    now_tz = utc_to_timezone(datetime.utcnow(), self.recipe.timezone)
    date_tz = str(now_tz.date())

    self.assertEqual(status['date_tz'], date_tz)
    self.assertEqual(len(status['tasks']), len(self.recipe.get_json()['tasks']))

    # includes all tasks in sequence including hours=[], normally #2 is skipped
    self.assertEqual(status['tasks'][0]['order'], 0)
    self.assertEqual(status['tasks'][1]['order'], 1)
    self.assertEqual(status['tasks'][2]['order'], 2)
    self.assertEqual(status['tasks'][3]['order'], 3)
    self.assertEqual(status['tasks'][4]['order'], 4)
    self.assertEqual(status['tasks'][5]['order'], 5)
    self.assertEqual(status['tasks'][6]['order'], 6)

  def test_canceled(self):
    self.recipe.force()
    status = self.recipe.get_status()

    self.assertFalse(status['tasks'][0]['done'])
    self.assertEqual(status['tasks'][0]['event'], 'JOB_PENDING')
    self.assertFalse(status['tasks'][1]['done'])
    self.assertEqual(status['tasks'][1]['event'], 'JOB_PENDING')
    self.assertFalse(status['tasks'][2]['done'])
    self.assertEqual(status['tasks'][2]['event'], 'JOB_PENDING')
    self.assertFalse(status['tasks'][3]['done'])
    self.assertEqual(status['tasks'][3]['event'], 'JOB_PENDING')
    self.assertFalse(status['tasks'][4]['done'])
    self.assertEqual(status['tasks'][4]['event'], 'JOB_PENDING')
    self.assertFalse(status['tasks'][5]['done'])
    self.assertEqual(status['tasks'][5]['event'], 'JOB_PENDING')
    self.assertFalse(status['tasks'][6]['done'])
    self.assertEqual(status['tasks'][6]['event'], 'JOB_PENDING')
    self.assertEqual(self.recipe.job_utm, self.recipe.get_job_utm(status))

    self.recipe.cancel()
    status = self.recipe.get_status()

    self.assertTrue(status['tasks'][0]['done'])
    self.assertEqual(status['tasks'][0]['event'], 'JOB_CANCEL')
    self.assertTrue(status['tasks'][1]['done'])
    self.assertEqual(status['tasks'][1]['event'], 'JOB_CANCEL')
    self.assertTrue(status['tasks'][2]['done'])
    self.assertEqual(status['tasks'][2]['event'], 'JOB_CANCEL')
    self.assertTrue(status['tasks'][3]['done'])
    self.assertEqual(status['tasks'][3]['event'], 'JOB_CANCEL')
    self.assertTrue(status['tasks'][4]['done'])
    self.assertEqual(status['tasks'][4]['event'], 'JOB_CANCEL')
    self.assertTrue(status['tasks'][5]['done'])
    self.assertEqual(status['tasks'][5]['event'], 'JOB_CANCEL')
    self.assertTrue(status['tasks'][6]['done'])
    self.assertEqual(status['tasks'][6]['event'], 'JOB_CANCEL')
    self.assertEqual(self.recipe.job_utm, self.recipe.get_job_utm(status))

  def test_hour_pulls(self):
    hour_tz = utc_to_timezone(datetime.utcnow(), self.recipe.timezone).hour

    if hour_tz == 23:
      print('SKIPPING test_hour_pulls, need 1 spare hour for test.')
    else:
      task = self.recipe.get_task()
      while task != None:
        self.recipe.set_task(task['script'], task['instance'], task['hour'],
                             'JOB_END', 'Some output.', '')
        task = self.recipe.get_task()

      status = self.recipe.get_status()

      self.assertTrue(
          all([(task['done'] and task['hour'] <= hour_tz) or
               (not task['done'] and task['hour'] > hour_tz)
               for task in status['tasks']]))

  def test_worker(self):

    # remove time dependency for this test, force all tasks
    self.recipe.force()
    status = self.recipe.get_status()

    for task in status['tasks']:

      ignore, job = worker_pull('SAMPLE_WORKER', jobs=1)
      self.assertEqual(len(job), 1)
      job = job[0]

      self.assertEqual(job['event'], 'JOB_PENDING')
      self.assertEqual(job['instance'], task['instance'])
      self.assertEqual(job['hour'], task['hour'])

      # job is not run through actual worker, so 'job' key will be missing, simulate it
      job['job'] = {
          'worker': 'SAMPLE_WORKER',
          'id': 'SAMPLE_JOB_UUID',
          'process': None,
          'utc': datetime.utcnow(),
      }

      worker_status(job['job']['worker'], job['recipe']['setup']['uuid'],
                    job['script'], job['instance'], job['hour'], 'JOB_END',
                    'Output is OK.', '')

      sleep(WORKER_LOOKBACK_EXPIRE / 1000)

    # after completing all tasks, check if they whole recipe is done
    self.recipe.refresh_from_db()
    assertRecipeDone(self, self.recipe)

    #self.assertGreater(self.recipe.job_utm, utc_milliseconds())

    #status = self.recipe.get_status()
    #self.assertTrue(all(task['event'] == 'JOB_END' for task in status['tasks']))

  def test_log(self):
    self.recipe.update()
    log = self.recipe.get_log()

    self.assertEqual(log['ago'], '1 Minute Ago')
    self.assertEqual(log['uid'], self.recipe.uid())
    self.assertEqual(log['percent'], 0)
    self.assertEqual(log['status'], 'NEW')

    self.recipe.force()
    log = self.recipe.get_log()

    self.assertEqual(log['ago'], '1 Minute Ago')
    self.assertEqual(log['uid'], self.recipe.uid())
    self.assertEqual(log['percent'], 0)
    self.assertEqual(log['status'], 'QUEUED')


class ManualTest(TestCase):

  def setUp(self):
    self.account = account_create()
    self.project = project_create()

    self.recipe = Recipe.objects.create(
        account=self.account,
        project=self.project,
        name='RECIPE_MANUAL',
        active=True,
        manual=True,
        week=[],
        hour=[],
        timezone='America/Los_Angeles',
        tasks=json.dumps([
            {
                'tag': 'manual',
                'values': {},
                'sequence': 1
            },
        ]),
    )
    self.recipe.update()

    now_tz = utc_to_timezone(datetime.utcnow(), self.recipe.timezone)

    self.recipe_done = Recipe.objects.create(
        account=self.account,
        project=self.project,
        name='RECIPE_MANUAL',
        active=True,
        manual=True,
        week=[],
        hour=[],
        timezone='America/Los_Angeles',
        tasks=json.dumps([
            {
                'tag': 'manual',
                'values': {},
                'sequence': 1
            },
        ]),
        job_status=json.dumps({
            'date_tz':
                str(now_tz.date()),
            'tasks': [{
                'instance': 1,
                'order': 0,
                'event': 'JOB_END',
                'utc': str(datetime.utcnow()),
                'script': 'manual',
                'hour': 0,
                'stdout': '',
                'stderr': '',
                'done': True
            }]
        }),
        worker_uid='TEST_WORKER',
        worker_utm=utc_milliseconds() - WORKER_LOOKBACK_EXPIRE)
    self.recipe_done.update()

  def test_done(self):
    self.assertIsNone(self.recipe_done.get_task())
    self.recipe_done.force()
    self.assertIsNotNone(self.recipe_done.get_task())

  def test_status(self):
    # without force manual tasks do not pull
    status = self.recipe.get_status()

    now_tz = utc_to_timezone(datetime.utcnow(), self.recipe.timezone)
    date_tz = str(now_tz.date())
    hour_tz = now_tz.hour

    self.assertEqual(status['date_tz'], date_tz)
    self.assertEqual(len(status['tasks']), 0)

    # force a run now on a manual task
    self.recipe.force()
    status = self.recipe.get_status()
    self.assertEqual(len(status['tasks']), 1)

  def test_worker(self):

    # no jobs
    ignore, job = worker_pull('SAMPLE_WORKER', jobs=0)
    self.assertEqual(len(job), 0)

    # manual mode ( without force always returns no tasks )
    ignore, job = worker_pull('SAMPLE_WORKER', jobs=1)
    self.assertEqual(len(job), 0)

    # advance time, since current jobs need to expire, artificially ping to keep out of queue
    sleep(WORKER_LOOKBACK_EXPIRE / 1000)

    # remove time dependency for this test, force all tasks in first recipe
    status = self.recipe.force()

    for task in status['tasks']:

      ignore, job = worker_pull('SAMPLE_WORKER', jobs=1)
      self.assertEqual(len(job), 1)
      job = job[0]

      self.assertEqual(job['event'], 'JOB_PENDING')
      self.assertEqual(job['instance'], task['instance'])
      self.assertEqual(job['hour'], task['hour'])

      # job is not run through actual worker, so 'job' key will be missing, simulate it
      job['job'] = {
          'worker': 'SAMPLE_WORKER',
          'id': 'SAMPLE_JOB_UUID',
          'process': None,
          'utc': datetime.utcnow(),
      }

      worker_status(job['job']['worker'], job['recipe']['setup']['uuid'],
                    job['script'], job['instance'], job['hour'], 'JOB_END',
                    'Output is OK.', '')

      sleep(WORKER_LOOKBACK_EXPIRE / 1000)

    # after completing all tasks, check if the whole recipe is done
    assertRecipeDone(self, self.recipe)


class RecipeViewTest(TestCase):

  def setUp(self):
    self.account = account_create()
    self.project = project_create()

    self.recipe_new = Recipe.objects.create(
        account=self.account,
        project=self.project,
        name='RECIPE_NEW',
        active=True,
        week=json.dumps(['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']),
        hour=json.dumps([0]),
        timezone='America/Los_Angeles',
        tasks=json.dumps([
            {
                'tag': 'hello',
                'values': {
                    'say_first': 'Hi Once',
                    'say_second': 'Hi Twice',
                    'sleep': 0
                },
                'sequence': 1
            },
        ]),
    )
    self.RECIPE_NEW = self.recipe_new.uid()

  def test_recipe_list(self):
    resp = self.client.get('/')
    self.assertEqual(resp.status_code, 200)

  def test_recipe_edit(self):
    resp = self.client.get('/recipe/edit/')
    self.assertEqual(resp.status_code, 302)

  def test_recipe_manual(self):
    resp = self.client.get('/recipe/edit/?manual=true')
    self.assertEqual(resp.status_code, 302)

  def test_recipe_start(self):
    resp = self.client.post('/recipe/start/',
                            {'reference': 'THISREFERENCEISINVALID'})
    self.assertEqual(resp.status_code, 404)
    self.assertEqual(resp.content, b'RECIPE NOT FOUND')

    resp = self.client.post('/recipe/start/',
                            {'reference': self.recipe_new.reference})
    self.assertEqual(resp.status_code, 200)
    self.assertEqual(resp.content, b'RECIPE STARTED')

    # start a recipe run to test interrupt
    ignore, jobs = worker_pull('SAMPLE_WORKER', 1)
    self.assertEqual(len(jobs), 1)

    resp = self.client.post('/recipe/start/',
                            {'reference': self.recipe_new.reference})
    self.assertEqual(resp.status_code, 200)
    self.assertEqual(resp.content, b'RECIPE INTERRUPTED')

  def test_recipe_stop(self):
    resp = self.client.post('/recipe/stop/',
                            {'reference': 'THISREFERENCEISINVALID'})
    self.assertEqual(resp.status_code, 404)
    self.assertEqual(resp.content, b'RECIPE NOT FOUND')

    resp = self.client.post('/recipe/stop/',
                            {'reference': self.recipe_new.reference})
    self.assertEqual(resp.status_code, 200)
    self.assertEqual(resp.content, b'RECIPE STOPPED')

    #self.recipe_new.refresh_from_db()
    #self.assertTrue(self.recipe_new.job_done)
    assertRecipeDone(self, self.recipe_new)

    # recipe is stopped and cannot be pulled
    ignore, jobs = worker_pull('SAMPLE_WORKER', 1)
    self.assertEqual(len(jobs), 0)

    # start a recipe run to test interrupt
    resp = self.client.post('/recipe/start/',
                            {'reference': self.recipe_new.reference})
    self.assertEqual(resp.status_code, 200)
    self.assertEqual(resp.content, b'RECIPE STARTED')

    #self.recipe_new.refresh_from_db()
    #self.assertFalse(self.recipe_new.job_done)
    assertRecipeNotDone(self, self.recipe_new)

    # wait for next worker cycle
    sleep(WORKER_LOOKBACK_EXPIRE / 1000)

    # recipe is started and can be pulled
    ignore, jobs = worker_pull('SAMPLE_WORKER', 1)
    self.assertEqual(len(jobs), 1)

    resp = self.client.post('/recipe/stop/',
                            {'reference': self.recipe_new.reference})
    self.assertEqual(resp.status_code, 200)
    self.assertEqual(resp.content, b'RECIPE INTERRUPTED')

    self.recipe_new.refresh_from_db()
    #self.assertTrue(self.recipe_new.job_done)
    assertRecipeDone(self, self.recipe_new)

    # wait for next worker cycle
    sleep(WORKER_LOOKBACK_EXPIRE / 1000)

    # recipe is stopped and cannot be pulled
    ignore, jobs = worker_pull('SAMPLE_WORKER', 1)
    self.assertEqual(len(jobs), 0)


class JobTest(TransactionTestCase):

  def setUp(self):
    self.account = account_create()
    self.project = project_create()

    now_tz = utc_to_timezone(datetime.utcnow(), 'America/Los_Angeles')
    status = {
        'date_tz':
            str(now_tz.date()),
        'tasks': [
            {
                'instance': 1,
                'order': 0,
                'event': 'JOB_PENDING',
                'utc': str(datetime.utcnow()),
                'script': 'hello',
                'hour': 0,
                'stdout': '',
                'stderr': '',
                'done': False
            },
            {
                'instance': 2,
                'order': 2,
                'event': 'JOB_PENDING',
                'utc': str(datetime.utcnow()),
                'script': 'hello',
                'hour': 0,
                'stdout': '',
                'stderr': '',
                'done': False
            },
        ]
    }

    self.job_new = Recipe.objects.create(
        account=self.account,
        project=self.project,
        name='RECIPE_NEW',
        active=True,
        week=json.dumps(['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']),
        hour=json.dumps([0]),
        timezone='America/Los_Angeles',
        tasks=json.dumps([
            {
                'tag': 'hello',
                'values': {
                    'say_first': 'Hi Once',
                    'say_second': 'Hi Twice',
                    'sleep': 0
                },
                'sequence': 1
            },
        ]),
    )
    self.RECIPE_NEW = self.job_new.uid()

    self.job_queued = Recipe.objects.create(
        account=self.account,
        project=self.project,
        name='RECIPE_QUEUED',
        active=True,
        week=json.dumps(['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']),
        hour=json.dumps([0]),
        timezone='America/Los_Angeles',
        tasks=json.dumps([
            {
                'tag': 'hello',
                'values': {
                    'say_first': 'Hi Once',
                    'say_second': 'Hi Twice',
                    'sleep': 0
                },
                'sequence': 1
            },
        ]),
        job_status=json.dumps(status.copy()),
        job_utm=test_job_utm())
    self.RECIPE_QUEUED = self.job_queued.uid()

    self.job_expired = Recipe.objects.create(
        account=self.account,
        project=self.project,
        name='RECIPE_EXPIRED',
        active=True,
        week=json.dumps(['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']),
        hour=json.dumps([0]),
        timezone='America/Los_Angeles',
        tasks=json.dumps([
            {
                'tag': 'hello',
                'values': {
                    'say_first': 'Hi Once',
                    'say_second': 'Hi Twice',
                    'sleep': 0
                },
                'sequence': 1
            },
        ]),
        job_status=json.dumps(status.copy()),
        job_utm=test_job_utm(),
        worker_uid='SAMPLE_WORKER',
        worker_utm=utc_milliseconds() - WORKER_LOOKBACK_EXPIRE)
    self.RECIPE_EXPIRED = self.job_expired.uid()

    self.job_running = Recipe.objects.create(
        account=self.account,
        project=self.project,
        name='RECIPE_RUNNING',
        active=True,
        week=json.dumps(['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']),
        hour=json.dumps([0]),
        timezone='America/Los_Angeles',
        tasks=json.dumps([{
            'tag': 'hello',
            'values': {
                'say_first': 'Hi Once',
                'say_second': 'Hi Twice',
                'sleep': 0
            },
            'sequence': 1
        }]),
        job_status=json.dumps(status.copy()),
        job_utm=test_job_utm(),
        worker_uid='OTHER_WORKER',
        worker_utm=utc_milliseconds())
    self.RECIPE_RUNNING = self.job_running.uid()

    self.job_paused = Recipe.objects.create(
        account=self.account,
        project=self.project,
        name='RECIPE_PAUSED',
        active=False,
        week=json.dumps(['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']),
        hour=json.dumps([0]),
        timezone='America/Los_Angeles',
        tasks=json.dumps([{
            'tag': 'hello',
            'values': {
                'say_first': 'Hi Once',
                'say_second': 'Hi Twice',
                'sleep': 0
            },
            'sequence': 1
        }]),
        job_status=json.dumps(status.copy()),
        job_utm=test_job_utm(),
        worker_uid='OTHER_WORKER',
        worker_utm=utc_milliseconds() - WORKER_LOOKBACK_EXPIRE)
    self.RECIPE_PAUSED = self.job_paused.uid()

    # paused so its not part of the normal flow ( unpause to use in test )
    self.job_error = Recipe.objects.create(
        account=self.account,
        project=self.project,
        name='RECIPE_ERROR',
        active=False,
        week=json.dumps(['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']),
        hour=json.dumps([0]),
        timezone='America/Los_Angeles',
        tasks=json.dumps([
            {
                'tag': 'hello',
                'values': {
                    'say_first': 'Hi Once',
                    'say_second': 'Hi Twice',
                    'sleep': 0,
                    'errror': 'An error is triggered.'
                },
                'sequence': 1
            },
        ]),
        job_status=json.dumps(status.copy()),
        job_utm=test_job_utm(),
        worker_uid='',
        worker_utm=0)
    self.RECIPE_ERROR = self.job_error.uid()

  def test_single_pulls(self):

    # first pull new task 1
    ignore, jobs = worker_pull('SAMPLE_WORKER', 1)
    self.assertEqual(len(jobs), 1)
    self.assertEqual(jobs[0]['recipe']['setup']['uuid'], self.RECIPE_QUEUED)
    self.assertEqual(jobs[0]['script'], 'hello')
    self.assertEqual(jobs[0]['instance'], 1)
    self.assertEqual(jobs[0]['hour'], 0)

    # second pull expired task 1
    ignore, jobs = worker_pull('SAMPLE_WORKER', 1)
    self.assertEqual(len(jobs), 1)
    self.assertEqual(jobs[0]['recipe']['setup']['uuid'], self.RECIPE_EXPIRED)
    self.assertEqual(jobs[0]['script'], 'hello')
    self.assertEqual(jobs[0]['instance'], 1)
    self.assertEqual(jobs[0]['hour'], 0)

    # third pull is blank since all recipes have been pulled from
    ignore, jobs = worker_pull('SAMPLE_WORKER', 1)
    self.assertEqual(len(jobs), 0)

    # expire all workers except OTHER_WORKER / RECIPE_RUNNING
    sleep(WORKER_LOOKBACK_EXPIRE / 1000)
    worker_ping('OTHER_WORKER', [self.RECIPE_RUNNING])

    # get oldest expired job first ( redo task since it never completes )
    ignore, jobs = worker_pull('SAMPLE_WORKER', 1)
    self.assertEqual(jobs[0]['recipe']['setup']['uuid'], self.RECIPE_QUEUED)
    self.assertEqual(jobs[0]['script'], 'hello')
    self.assertEqual(jobs[0]['instance'], 1)
    self.assertEqual(jobs[0]['hour'], 0)

  def test_multiple_pulls(self):

    # pull all jobs at once
    ignore, jobs = worker_pull('TEST_WORKER', 5)
    self.assertEqual(len(jobs), 2)

    self.assertEqual(jobs[0]['recipe']['setup']['uuid'], self.RECIPE_QUEUED)
    self.assertEqual(jobs[0]['script'], 'hello')
    self.assertEqual(jobs[0]['instance'], 1)
    self.assertEqual(jobs[0]['hour'], 0)

    self.assertEqual(jobs[1]['recipe']['setup']['uuid'], self.RECIPE_EXPIRED)
    self.assertEqual(jobs[1]['script'], 'hello')
    self.assertEqual(jobs[1]['instance'], 1)
    self.assertEqual(jobs[1]['hour'], 0)

  def test_manager(self):

    management.call_command(
        'job_worker',
        worker='TEST_WORKER',
        jobs=5,
        timeout=60 * 60 * 1,
        verbose=True,
        test=True)

    jobs = Recipe.objects.filter(worker_uid='TEST_WORKER')
    self.assertEqual(len(jobs), 2)
    status = jobs[0].get_status()
    self.assertEqual(len(status['tasks']), 2)
    self.assertEqual(status['tasks'][0]['script'], 'hello')
    self.assertEqual(status['tasks'][0]['instance'], 1)
    self.assertEqual(status['tasks'][0]['hour'], 0)
    self.assertEqual(status['tasks'][0]['event'], 'JOB_END')
    self.assertEqual(status['tasks'][1]['script'], 'hello')
    self.assertEqual(status['tasks'][1]['instance'], 2)
    self.assertEqual(status['tasks'][1]['hour'], 0)
    self.assertEqual(status['tasks'][1]['event'], 'JOB_PENDING')

    # advance time, since current jobs need to expire, artificially ping to keep out of queue
    sleep(WORKER_LOOKBACK_EXPIRE / 1000)
    worker_ping('OTHER_WORKER', [self.RECIPE_RUNNING])

    # second loop through manager
    management.call_command(
        'job_worker',
        worker='TEST_WORKER',
        jobs=5,
        timeout=60 * 60 * 1,
        verbose=True,
        test=True)

    jobs = Recipe.objects.filter(worker_uid='TEST_WORKER')
    self.assertEqual(len(jobs), 2)
    status = jobs[0].get_status()
    self.assertEqual(len(status['tasks']), 2)
    self.assertEqual(status['tasks'][0]['script'], 'hello')
    self.assertEqual(status['tasks'][0]['instance'], 1)
    self.assertEqual(status['tasks'][0]['hour'], 0)
    self.assertEqual(status['tasks'][0]['event'], 'JOB_END')
    self.assertEqual(status['tasks'][1]['script'], 'hello')
    self.assertEqual(status['tasks'][1]['instance'], 2)
    self.assertEqual(status['tasks'][1]['hour'], 0)
    self.assertEqual(status['tasks'][1]['event'], 'JOB_END')

    # jobs were also marked as complete
    #self.assertTrue(jobs[0].job_done)
    #self.assertTrue(jobs[1].job_done)
    assertRecipeDone(self, jobs[0])
    assertRecipeDone(self, jobs[1])

    # advance time, since current jobs need to expire, artificially ping to keep out of queue
    sleep(WORKER_LOOKBACK_EXPIRE / 1000)
    worker_ping('OTHER_WORKER', [self.RECIPE_RUNNING])

    # all jobs either run by other workers or done
    ignore, jobs = worker_pull('TEST_WORKER', 5)
    self.assertEqual(len(jobs), 0)


class JobErrorTest(TransactionTestCase):

  def setUp(self):
    self.account = account_create()
    self.project = project_create()

    now_tz = utc_to_timezone(datetime.utcnow(), 'America/Los_Angeles')

    self.recipe = Recipe.objects.create(
        account=self.account,
        project=self.project,
        name='RECIPE_NEW',
        active=True,
        week=json.dumps(['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']),
        hour=json.dumps([0]),
        timezone='America/Los_Angeles',
        tasks=json.dumps([
            {
                'tag': 'hello',
                'values': {
                    'error': 'Triggered the error mechanic on purpose.'
                },
                'sequence': 1
            },
        ]),
        job_status=json.dumps({
            'date_tz':
                str(now_tz.date()),
            'tasks': [
                {
                    'instance': 1,
                    'order': 0,
                    'event': 'JOB_PENDING',
                    'utc': str(datetime.utcnow()),
                    'script': 'hello',
                    'hour': 0,
                    'stdout': '',
                    'stderr': '',
                    'done': False
                },
                {
                    'instance': 2,
                    'order': 2,
                    'event': 'JOB_PENDING',
                    'utc': str(datetime.utcnow()),
                    'script': 'hello',
                    'hour': 0,
                    'stdout': '',
                    'stderr': '',
                    'done': False
                },
            ]
        }),
        job_utm=test_job_utm())

  def test_manager_error(self):

    management.call_command(
        'job_worker',
        worker='TEST_WORKER',
        jobs=5,
        timeout=60 * 60 * 1,
        verbose=True,
        test=True)

    jobs = Recipe.objects.filter(worker_uid='TEST_WORKER')
    self.assertEqual(len(jobs), 1)
    status = jobs[0].get_status()
    self.assertEqual(len(status['tasks']), 2)
    self.assertEqual(status['tasks'][0]['script'], 'hello')
    self.assertEqual(status['tasks'][0]['instance'], 1)
    self.assertEqual(status['tasks'][0]['hour'], 0)
    self.assertEqual(status['tasks'][0]['event'], 'JOB_ERROR')
    self.assertIn('PROJECT TASK SAY: Hello Once', status['tasks'][0]['stdout'])
    self.assertIn('Exception: Triggered the error mechanic on purpose.',
                  status['tasks'][0]['stderr'])

    self.assertEqual(status['tasks'][1]['script'], 'hello')
    self.assertEqual(status['tasks'][1]['instance'], 2)
    self.assertEqual(status['tasks'][1]['hour'], 0)
    self.assertEqual(status['tasks'][1]['event'], 'JOB_PENDING')
    self.assertEqual('', status['tasks'][1]['stdout'])
    self.assertEqual('', status['tasks'][1]['stderr'])

    # advance time, since current jobs need to expire, artificially ping to keep out of queue
    sleep(WORKER_LOOKBACK_EXPIRE / 1000)

    # second loop through manager
    management.call_command(
        'job_worker',
        worker='TEST_WORKER',
        jobs=5,
        timeout=60 * 60 * 1,
        verbose=True,
        test=True)

    jobs = Recipe.objects.filter(worker_uid='TEST_WORKER')
    self.assertEqual(len(jobs), 1)
    status = jobs[0].get_status()
    self.assertEqual(len(status['tasks']), 2)
    self.assertEqual(status['tasks'][0]['script'], 'hello')
    self.assertEqual(status['tasks'][0]['instance'], 1)
    self.assertEqual(status['tasks'][0]['hour'], 0)
    self.assertEqual(status['tasks'][0]['event'], 'JOB_ERROR')
    self.assertIn('PROJECT TASK SAY: Hello Once', status['tasks'][0]['stdout'])
    self.assertIn('Exception: Triggered the error mechanic on purpose.',
                  status['tasks'][0]['stderr'])
    self.assertEqual(status['tasks'][1]['script'], 'hello')
    self.assertEqual(status['tasks'][1]['instance'], 2)
    self.assertEqual(status['tasks'][1]['hour'], 0)
    self.assertEqual(status['tasks'][1]['event'], 'JOB_END')
    self.assertIn('PROJECT TASK SAY: Hello Twice', status['tasks'][1]['stdout'])
    self.assertEqual('', status['tasks'][1]['stderr'])

    # check if recipe is removed from worker lookup ( job_done=True )
    assertRecipeDone(self, self.recipe)


class JobTimeoutTest(TransactionTestCase):

  def setUp(self):
    self.account = account_create()
    self.project = project_create()

    now_tz = utc_to_timezone(datetime.utcnow(), 'America/Los_Angeles')

    self.recipe = Recipe.objects.create(
        account=self.account,
        project=self.project,
        name='RECIPE_NEW',
        active=True,
        week=json.dumps(['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']),
        hour=json.dumps([0]),
        timezone='America/Los_Angeles',
        tasks=json.dumps([
            {
                'tag': 'hello',
                'values': {
                    'sleep': 15
                },  # seconds
                'sequence': 1
            },
        ]),
        job_status=json.dumps({
            'date_tz':
                str(now_tz.date()),
            'tasks': [
                {
                    'instance': 1,
                    'order': 0,
                    'event': 'JOB_PENDING',
                    'utc': str(datetime.utcnow()),
                    'script': 'hello',
                    'hour': 0,
                    'stdout': '',
                    'stderr': '',
                    'done': False
                },
                {
                    'instance': 2,
                    'order': 2,
                    'event': 'JOB_PENDING',
                    'utc': str(datetime.utcnow()),
                    'script': 'hello',
                    'hour': 0,
                    'stdout': '',
                    'stderr': '',
                    'done': False
                },
            ]
        }),
        job_utm=test_job_utm())

  def test_manager_timeout(self):

    # first loop through manager ( use short timeout )
    management.call_command(
        'job_worker',
        worker='TEST_WORKER',
        jobs=5,
        timeout=5,
        verbose=True,
        test=True)

    jobs = Recipe.objects.filter(worker_uid='TEST_WORKER')
    self.assertEqual(len(jobs), 1)
    status = jobs[0].get_status()
    self.assertEqual(len(status['tasks']), 2)
    self.assertEqual(status['tasks'][0]['script'], 'hello')
    self.assertEqual(status['tasks'][0]['instance'], 1)
    self.assertEqual(status['tasks'][0]['hour'], 0)
    self.assertEqual(status['tasks'][0]['event'], 'JOB_TIMEOUT')
    self.assertEqual(status['tasks'][1]['script'], 'hello')
    self.assertEqual(status['tasks'][1]['instance'], 2)
    self.assertEqual(status['tasks'][1]['hour'], 0)
    self.assertEqual(status['tasks'][1]['event'], 'JOB_PENDING')

    # advance time, since current jobs need to expire, artificially ping to keep out of queue
    sleep(WORKER_LOOKBACK_EXPIRE / 1000)

    # second loop through manager ( use normal timeout )
    management.call_command(
        'job_worker',
        worker='TEST_WORKER',
        jobs=5,
        timeout=60 * 60 * 1,
        verbose=True,
        test=True)

    jobs = Recipe.objects.filter(worker_uid='TEST_WORKER')
    self.assertEqual(len(jobs), 1)
    status = jobs[0].get_status()
    self.assertEqual(len(status['tasks']), 2)
    self.assertEqual(status['tasks'][0]['script'], 'hello')
    self.assertEqual(status['tasks'][0]['instance'], 1)
    self.assertEqual(status['tasks'][0]['hour'], 0)
    self.assertEqual(status['tasks'][0]['event'], 'JOB_TIMEOUT')
    self.assertEqual(status['tasks'][1]['script'], 'hello')
    self.assertEqual(status['tasks'][1]['instance'], 2)
    self.assertEqual(status['tasks'][1]['hour'], 0)
    self.assertEqual(status['tasks'][1]['event'], 'JOB_END')

    # check if recipe is removed from worker lookup ( job_done=True )
    assertRecipeDone(self, self.recipe)


class JobDayTest(TransactionTestCase):

  def setUp(self):
    self.account = account_create()
    self.project = project_create()

    now_tz = utc_to_timezone(datetime.utcnow(), 'America/Los_Angeles')
    today_tz = now_tz.strftime('%a')
    yesterday_tz = (now_tz - timedelta(days=1)).strftime('%a')
    tomorrow_tz = (now_tz + timedelta(days=1)).strftime('%a')

    self.recipe_today = Recipe.objects.create(
        account=self.account,
        project=self.project,
        name='RECIPE_TODAY',
        active=True,
        week=json.dumps([today_tz]),
        hour=json.dumps([0]),
        timezone='America/Los_Angeles',
        tasks=json.dumps([
            {
                'tag': 'hello',
                'values': {},
                'sequence': 1
            },
        ]),
        job_status=json.dumps({
            'date_tz':
                str(now_tz.date()),
            'tasks': [{
                'instance': 1,
                'order': 0,
                'event': 'JOB_PENDING',
                'utc': str(datetime.utcnow()),
                'script': 'hello',
                'hour': 0,
                'stdout': '',
                'stderr': '',
                'done': False
            }, {
                'instance': 2,
                'order': 0,
                'event': 'JOB_PENDING',
                'utc': str(datetime.utcnow()),
                'script': 'hello',
                'hour': 0,
                'stdout': '',
                'stderr': '',
                'done': False
            }]
        }),
        worker_uid='OTHER_WORKER',
        worker_utm=utc_milliseconds() - WORKER_LOOKBACK_EXPIRE,
        job_utm=test_job_utm())

    self.recipe_not_today = Recipe.objects.create(
        account=self.account,
        project=self.project,
        name='RECIPE_NOT_TODAY',
        active=True,
        week=json.dumps([yesterday_tz, tomorrow_tz]),
        hour=json.dumps([0]),
        timezone='America/Los_Angeles',
        tasks=json.dumps([
            {
                'tag': 'hello',
                'values': {},
                'sequence': 1
            },
        ]),
        job_status=json.dumps({
            'date_tz':
                str(now_tz.date() - timedelta(days=1)),
            'tasks': [{
                'instance': 1,
                'order': 0,
                'event': 'JOB_END',
                'utc': str(datetime.utcnow() - timedelta(days=1)),
                'script': 'hello',
                'hour': 0,
                'stdout': '',
                'stderr': '',
                'done': True
            }, {
                'instance': 2,
                'order': 0,
                'event': 'JOB_END',
                'utc': str(datetime.utcnow() - timedelta(days=1)),
                'script': 'hello',
                'hour': 0,
                'stdout': '',
                'stderr': '',
                'done': True
            }]
        }),
        worker_uid='OTHER_WORKER',
        worker_utm=utc_milliseconds() - WORKER_LOOKBACK_EXPIRE,
        job_utm=test_job_utm(0, 1))

  def test_job_day(self):
    # test low level pull ( no worker involved )
    self.assertIsNotNone(self.recipe_today.get_task())
    self.assertIsNone(self.recipe_not_today.get_task())

    assertRecipeNotDone(self, self.recipe_today)
    assertRecipeDone(self, self.recipe_not_today)

    # first loop through manager ( use short timeout )
    management.call_command(
        'job_worker',
        worker='TEST_WORKER',
        jobs=5,
        timeout=60,
        verbose=True,
        test=True)

    recipes = Recipe.objects.filter(worker_uid='TEST_WORKER')
    self.assertEqual(len(recipes), 1)
    self.assertEqual(recipes[0].name, 'RECIPE_TODAY')


class JobCancelTest(TransactionTestCase):

  def setUp(self):
    self.account = account_create()
    self.project = project_create()

    now_tz = utc_to_timezone(datetime.utcnow(), 'America/Los_Angeles')

    self.recipe = Recipe.objects.create(
        account=self.account,
        project=self.project,
        name='RECIPE_NEW',
        active=True,
        week=json.dumps(['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']),
        hour=json.dumps([0]),
        timezone='America/Los_Angeles',
        tasks=json.dumps([
            {
                'tag': 'hello',
                'values': {
                    'sleep': 10
                },  # seconds
                'sequence': 1
            },
        ]),
        job_status=json.dumps({
            'date_tz':
                str(now_tz.date()),
            'tasks': [{
                'instance': 1,
                'order': 0,
                'event': 'JOB_END',
                'utc': str(datetime.utcnow() - timedelta(days=1)),
                'script': 'hello',
                'hour': 0,
                'stdout': '',
                'stderr': '',
                'done': False
            }, {
                'instance': 2,
                'order': 0,
                'event': 'JOB_END',
                'utc': str(datetime.utcnow() - timedelta(days=1)),
                'script': 'hello',
                'hour': 0,
                'stdout': '',
                'stderr': '',
                'done': False
            }]
        }),
        job_utm=test_job_utm())

    self.workers = Workers('TEST_WORKER', 5, 60)

  def tearDown(self):
    # if your threads are locking in the test, comment out the line below, that will allow the error to be printed before deadlocking
    self.workers.shutdown()

  def test_job_cancel(self):

    self.workers.pull()
    self.assertEqual(len(self.workers.jobs), 1)

    self.workers.poll()
    job = self.workers.jobs[0]
    self.assertEqual(job['job']['worker'], 'TEST_WORKER')
    self.assertEqual(job['recipe']['setup']['uuid'], self.recipe.uid())
    self.assertEqual(job['script'], 'hello')
    self.assertEqual(job['instance'], 1)
    self.assertIsNone(job['job']['process'].poll())

    self.recipe.cancel()

    self.assertEqual(len(self.workers.jobs), 1)
    self.workers.pull()
    self.assertEqual(len(self.workers.jobs), 0)

    assertRecipeDone(self, self.recipe)
    status = self.recipe.get_status()
    self.assertEqual(status['tasks'][0]['event'], 'JOB_CANCEL')

  def test_job_cancel_early(self):

    self.workers.pull()
    self.assertEqual(len(self.workers.jobs), 1)

    self.recipe.cancel()

    self.assertEqual(len(self.workers.jobs), 1)
    self.workers.pull()
    self.assertEqual(len(self.workers.jobs), 0)

    assertRecipeDone(self, self.recipe)
    status = self.recipe.get_status()
    self.assertEqual(status['tasks'][0]['event'], 'JOB_CANCEL')


class WorkerTest(TransactionTestCase):

  def setUp(self):
    self.account = account_create()
    self.project = project_create()

    now_tz = utc_to_timezone(datetime.utcnow(), 'America/Los_Angeles')

    self.job_done = Recipe.objects.create(
        account=self.account,
        project=self.project,
        name='RECIPE_DONE',
        active=True,
        week=json.dumps(['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']),
        hour=json.dumps([0]),
        timezone='America/Los_Angeles',
        tasks=json.dumps([
            {
                'tag': 'hello',
                'values': {
                    'say_first': 'Hi Once',
                    'say_second': 'Hi Twice',
                    'sleep': 0
                },
                'sequence': 1
            },
        ]),
        job_status=json.dumps({
            'date_tz':
                str(now_tz.date()),
            'tasks': [{
                'instance': 1,
                'order': 0,
                'event': 'JOB_END',
                'utc': str(datetime.utcnow()),
                'script': 'hello',
                'hour': now_tz.hour,
                'stdout': '',
                'stderr': '',
                'done': True
            }, {
                'instance': 2,
                'order': 0,
                'event': 'JOB_END',
                'utc': str(datetime.utcnow()),
                'script': 'hello',
                'hour': now_tz.hour,
                'stdout': '',
                'stderr': '',
                'done': True
            }]
        }),
        worker_uid='SAMPLE_WORKER',
        worker_utm=utc_milliseconds() - WORKER_LOOKBACK_EXPIRE,
        job_utm=test_job_utm(0, 1))
    self.RECIPE_DONE = self.job_done.uid()

  def test_ui_autoscale(self):
    resp = self.client.get('/recipe/autoscale/')
    self.assertEqual(resp.status_code, 200)

  def test_worker_downscale(self):
    self.assertIsNone(worker_downscale())

  def test_worker_upscale_zero_jobs(self):
    self.assertJSONEqual(
        autoscale('TEST').content, {
            'jobs': 0,
            'workers': {
                'jobs': settings.WORKER_JOBS,
                'max': settings.WORKER_MAX,
                'existing': 3,
                'required': 0
            }
        })

  def test_worker_upscale_one_jobs(self):
    self.job_done.force()
    self.assertJSONEqual(
        autoscale('TEST').content, {
            'jobs': 1,
            'workers': {
                'jobs': settings.WORKER_JOBS,
                'max': settings.WORKER_MAX,
                'existing': 3,
                'required': 1
            }
        })

  def test_worker_upscale_many_jobs(self):
    for i in range(0, 7 * settings.WORKER_JOBS):
      self.job_new = Recipe.objects.create(
          account=self.account,
          project=self.project,
          name='RECIPE_NEW_%d' % i,
          active=True,
          week=json.dumps(['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']),
          hour=json.dumps([0]),
          timezone='America/Los_Angeles',
          tasks=json.dumps([
              {
                  'tag': 'hello',
                  'values': {
                      'say_first': 'Hi Once',
                      'say_second': 'Hi Twice',
                      'sleep': 0
                  },
                  'sequence': 1
              },
          ]),
      )
      self.job_new.force()

    self.assertJSONEqual(
        autoscale('TEST').content, {
            'jobs': 7 * settings.WORKER_JOBS,
            'workers': {
                'jobs': settings.WORKER_JOBS,
                'max': settings.WORKER_MAX,
                'existing': 3,
                'required': 7
            }
        })

  def test_worker_upscale_limit(self):
    for i in range(0, 70 * settings.WORKER_JOBS):
      self.job_new = Recipe.objects.create(
          account=self.account,
          project=self.project,
          name='RECIPE_NEW_%d' % i,
          active=True,
          week=json.dumps(['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']),
          hour=json.dumps([0]),
          timezone='America/Los_Angeles',
          tasks=json.dumps([
              {
                  'tag': 'hello',
                  'values': {
                      'say_first': 'Hi Once',
                      'say_second': 'Hi Twice',
                      'sleep': 0
                  },
                  'sequence': 1
              },
          ]),
      )
      self.job_new.force()

    self.assertJSONEqual(
        autoscale('TEST').content, {
            'jobs': 70 * settings.WORKER_JOBS,
            'workers': {
                'jobs': settings.WORKER_JOBS,
                'max': settings.WORKER_MAX,
                'existing': 3,
                'required': settings.WORKER_MAX
            }
        })



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

Bulk Editor For DV360

Allows bulk editing DV360 through Sheets and BigQuery.

  - Select <b>Load Partners</b> and run the recipe. A Sheet called DV_Sheet_UNDEFINED will be created.
  - Fill in <i>Partners</i> tab filter.
  - Run <b>Load Advertisers</b>.
  - Fill in <i>Advertisers</i> tab filter.
  - Run <b>Load Insertion Orders</b> or <b>Load Line Items</b> or <b>Load Creatives</b>.
  - Fill in changes on all tabs with colored fields.
  - Select <i>Audit</i> and run the recipe.
  - Check the <b>Audit</b> and <b>Preview</b> tabs.
  - Select <i>Patch</i> and run the recipe.
  - Check the <b>Success</b> and <b>Error</b> tabs.
  - Patch can be run multiple times.
  - Patch ONLY changes edited fields.

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'auth_dv': 'user',  # Credentials used for dv.
  'auth_sheet': 'user',  # Credentials used for sheet.
  'auth_bigquery': 'service',  # Credentials used for bigquery.
  'recipe_slug': '',  # Name of Google BigQuery dataset to create.
  'command': 'Load Partners',  # Action to take.
}

RECIPE = {
  'setup': {
    'day': [
    ],
    'hour': [
    ]
  },
  'tasks': [
    {
      'dataset': {
        '__comment__': 'Ensure dataset exists.',
        'auth': {
          'field': {
            'name': 'auth_bigquery',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'dataset': {
          'field': {
            'name': 'recipe_slug',
            'prefix': 'DV_Sheet_',
            'kind': 'string',
            'order': 2,
            'default': '',
            'description': 'Name of Google BigQuery dataset to create.'
          }
        }
      }
    },
    {
      'drive': {
        '__comment__': 'Copy the default template to sheet with the recipe name',
        'auth': {
          'field': {
            'name': 'auth_sheet',
            'kind': 'authentication',
            'order': 1,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'copy': {
          'source': 'https://docs.google.com/spreadsheets/d/18G6cGo4j5SsY08H8P53R22D_Pm6m-zkE6APd3EDLf2c/',
          'destination': {
            'field': {
              'name': 'recipe_slug',
              'prefix': 'DV Sheet ',
              'kind': 'string',
              'order': 3,
              'default': '',
              'description': 'Name of Google Sheet to create.'
            }
          }
        }
      }
    },
    {
      'dv_editor': {
        '__comment': 'Depending on users choice, execute a different part of the solution.',
        'auth_dv': {
          'field': {
            'name': 'auth_dv',
            'kind': 'authentication',
            'order': 1,
            'default': 'user',
            'description': 'Credentials used for dv.'
          }
        },
        'auth_sheets': {
          'field': {
            'name': 'auth_sheet',
            'kind': 'authentication',
            'order': 2,
            'default': 'user',
            'description': 'Credentials used for sheet.'
          }
        },
        'auth_bigquery': {
          'field': {
            'name': 'auth_bigquery',
            'kind': 'authentication',
            'order': 3,
            'default': 'service',
            'description': 'Credentials used for bigquery.'
          }
        },
        'sheet': {
          'field': {
            'name': 'recipe_slug',
            'prefix': 'DV Sheet ',
            'kind': 'string',
            'order': 4,
            'default': '',
            'description': 'Name of Google Sheet to create.'
          }
        },
        'dataset': {
          'field': {
            'name': 'recipe_slug',
            'prefix': 'DV_Sheet_',
            'kind': 'string',
            'order': 5,
            'default': '',
            'description': 'Name of Google BigQuery dataset to create.'
          }
        },
        'command': {
          'field': {
            'name': 'command',
            'kind': 'choice',
            'choices': [
              'Clear Partners',
              'Clear Advertisers',
              'Clear Campaigns',
              'Clear Creatives',
              'Clear Insertion Orders',
              'Clear Line Items',
              'Clear Preview',
              'Clear Patch',
              'Load Partners',
              'Load Advertisers',
              'Load Campaigns',
              'Load Creatives',
              'Load Insertion Orders',
              'Load Line Items',
              'Preview',
              'Patch'
            ],
            'order': 6,
            'default': 'Load Partners',
            'description': 'Action to take.'
          }
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('dv360_editor', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Handles creation and updates of landing pages."""

from starthinker.task.traffic.dao import BaseDAO
from starthinker.task.traffic.feed import FieldMap


class LandingPageDAO(BaseDAO):
  """Landing page data access object.

  Inherits from BaseDAO and implements landing page specific logic for creating
  and
  updating landing pages.
  """

  def __init__(self, auth, profile_id, is_admin):
    """Initializes LandingPageDAO with profile id and authentication scheme."""

    super(LandingPageDAO, self).__init__(auth, profile_id, is_admin)

    self._id_field = FieldMap.CAMPAIGN_LANDING_PAGE_ID
    self._search_field = FieldMap.CAMPAIGN_LANDING_PAGE_NAME
    self._list_name = 'landingPages'
    self._entity = 'LANDING_PAGE'
    self._parent_filter_name = 'advertiserIds'
    self._parent_filter_field_name = FieldMap.ADVERTISER_ID
    self._parent_dao = None

  def _api(self, iterate=False):
    """Returns an DCM API instance for this DAO."""
    return super(LandingPageDAO, self)._api(iterate).advertiserLandingPages()

  def _process_update(self, item, feed_item):
    """Updates an landing page based on the values from the feed.

    Args:
      item: Object representing the landing page to be updated, this object is
        updated directly.
      feed_item: Feed item representing landing page values from the Bulkdozer
        feed.
    """

    item['name'] = feed_item.get(FieldMap.CAMPAIGN_LANDING_PAGE_NAME, None)
    item['url'] = feed_item.get(FieldMap.CAMPAIGN_LANDING_PAGE_URL, None)

  def _process_new(self, feed_item):
    """Creates a new landing page DCM object from a feed item representing a landing page from the Bulkdozer feed.

    This function simply creates the object to be inserted later by the BaseDAO
    object.

    Args:
      feed_item: Feed item representing the landing page from the Bulkdozer
        feed.

    Returns:
      An landing page object ready to be inserted in DCM through the API.

    """
    return {
        'name': feed_item.get(FieldMap.CAMPAIGN_LANDING_PAGE_NAME, None),
        'url': feed_item.get(FieldMap.CAMPAIGN_LANDING_PAGE_URL, None),
        'advertiserId': feed_item.get(FieldMap.ADVERTISER_ID, None)
    }
